"""
Network ACL
===========

Manage the firewall configuration on the network device managed through NAPALM.
The firewall configuration is generated by Capirca_.

.. _Capirca: https://github.com/google/capirca

.. versionadded:: 2017.7.0

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:depends:    capirca, napalm
:platform:   unix

Dependencies
------------

Capirca
~~~~~~~

To install Capirca, execute: ``pip install capirca``.

NAPALM
~~~~~~

To be able to load configuration on network devices,
it requires NAPALM_ library to be installed:  ``pip install napalm``.
Please check Installation_ for complete details.

.. _NAPALM: https://napalm.readthedocs.io
.. _Installation: https://napalm.readthedocs.io/en/latest/installation/index.html
"""

import logging

import salt.utils.napalm

log = logging.getLogger(__file__)

try:
    # pylint: disable=W0611
    import capirca
    import capirca.aclgen
    import capirca.lib.policy
    import capirca.lib.aclgenerator

    HAS_CAPIRCA = True
    # pylint: enable=W0611
except ImportError:
    HAS_CAPIRCA = False


# ------------------------------------------------------------------------------
# state properties
# ------------------------------------------------------------------------------

__virtualname__ = "netacl"

# ------------------------------------------------------------------------------
# global variables
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# property functions
# ------------------------------------------------------------------------------


def __virtual__():
    """
    This module requires both NAPALM and Capirca.
    """
    if HAS_CAPIRCA and salt.utils.napalm.virtual(__opts__, __virtualname__, __file__):
        return __virtualname__
    else:
        return (
            False,
            "The netacl state cannot be loaded: Please install capirca and napalm.",
        )


# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def term(
    name,
    filter_name,
    term_name,
    filter_options=None,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=False,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
    test=False,
    commit=True,
    debug=False,
    source_service=None,
    destination_service=None,
    **term_fields
):
    """
    Manage the configuration of a specific policy term.

    filter_name
        The name of the policy filter.

    term_name
        The name of the term.

    filter_options
        Additional filter options. These options are platform-specific.
        See the complete list of options_.

        .. _options: https://github.com/google/capirca/wiki/Policy-format#header-section

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``False``
        Merge the CLI variables with the pillar. Default: ``False``.

        The properties specified through the state arguments have higher priority than the pillar.

    revision_id
        Add a comment in the term config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the term configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return the changes.
        Default: ``False`` and will commit the changes on the device.

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    source_service
        A special service to choose from. This is a helper so the user is able to
        select a source just using the name, instead of specifying a source_port and protocol.

        As this module is available on Unix platforms only,
        it reads the IANA_ port assignment from /etc/services.

        If the user requires additional shortcuts to be referenced, they can add entries under /etc/services,
        which can be managed using the :mod:`file state <salt.states.file>`.

        .. _IANA: http://www.iana.org/assignments/port-numbers

    destination_service
        A special service to choose from. This is a helper so the user is able to
        select a source just using the name, instead of specifying a destination_port and protocol.
        Allows the same options as ``source_service``.

    term_fields
        Term attributes. To see what fields are supported, please consult the
        list of supported keywords_. Some platforms have few other optional_
        keywords.

        .. _keywords: https://github.com/google/capirca/wiki/Policy-format#keywords
        .. _optional: https://github.com/google/capirca/wiki/Policy-format#optionally-supported-keywords

    .. note::
        The following fields are accepted:

        - action
        - address
        - address_exclude
        - comment
        - counter
        - expiration
        - destination_address
        - destination_address_exclude
        - destination_port
        - destination_prefix
        - forwarding_class
        - forwarding_class_except
        - logging
        - log_name
        - loss_priority
        - option
        - policer
        - port
        - precedence
        - principals
        - protocol
        - protocol_except
        - qos
        - pan_application
        - routing_instance
        - source_address
        - source_address_exclude
        - source_port
        - source_prefix
        - verbatim
        - packet_length
        - fragment_offset
        - hop_limit
        - icmp_type
        - ether_type
        - traffic_class_count
        - traffic_type
        - translated
        - dscp_set
        - dscp_match
        - dscp_except
        - next_ip
        - flexible_match_range
        - source_prefix_except
        - destination_prefix_except
        - vpn
        - source_tag
        - destination_tag
        - source_interface
        - destination_interface
        - flattened
        - flattened_addr
        - flattened_saddr
        - flattened_daddr
        - priority

    .. note::
        The following fields can be also a single value and a list of values:

        - action
        - address
        - address_exclude
        - comment
        - destination_address
        - destination_address_exclude
        - destination_port
        - destination_prefix
        - forwarding_class
        - forwarding_class_except
        - logging
        - option
        - port
        - precedence
        - principals
        - protocol
        - protocol_except
        - pan_application
        - source_address
        - source_address_exclude
        - source_port
        - source_prefix
        - verbatim
        - icmp_type
        - ether_type
        - traffic_type
        - dscp_match
        - dscp_except
        - flexible_match_range
        - source_prefix_except
        - destination_prefix_except
        - source_tag
        - destination_tag
        - source_service
        - destination_service

        Example: ``destination_address`` can be either defined as:

        .. code-block:: yaml

            destination_address: 172.17.17.1/24

        or as a list of destination IP addresses:

        .. code-block:: yaml

            destination_address:
                - 172.17.17.1/24
                - 172.17.19.1/24

        or a list of services to be matched:

        .. code-block:: yaml

            source_service:
                - ntp
                - snmp
                - ldap
                - bgpd

    .. note::
        The port fields ``source_port`` and ``destination_port`` can be used as
        above to select either a single value, either a list of values, but
        also they can select port ranges. Example:

        .. code-block:: yaml

            source_port:
                - [1000, 2000]
                - [3000, 4000]

        With the configuration above, the user is able to select the 1000-2000 and 3000-4000 source port ranges.

    CLI Example:

    .. code-block:: bash

        salt 'edge01.bjm01' state.sls router.acl

    Output Example:

    .. code-block:: text

        edge01.bjm01:
        ----------
                  ID: update_icmp_first_term
            Function: netacl.term
              Result: None
             Comment: Testing mode: Configuration discarded.
             Started: 12:49:09.174179
            Duration: 5751.882 ms
             Changes:
                      ----------
                      diff:
                          [edit firewall]
                          +    family inet {
                          +        /*
                          +         ** $Id: update_icmp_first_term $
                          +         ** $Date: 2017/02/30 $
                          +         **
                          +         */
                          +        filter block-icmp {
                          +            term first-term {
                          +                from {
                          +                    protocol icmp;
                          +                }
                          +                then {
                          +                    reject;
                          +                }
                          +            }
                          +        }
                          +    }

        Summary for edge01.bjm01
        ------------
        Succeeded: 1 (unchanged=1, changed=1)
        Failed:    0
        ------------
        Total states run:     1
        Total run time:   5.752 s

    Pillar example:

    .. code-block:: yaml

        firewall:
          - block-icmp:
              terms:
                - first-term:
                    protocol:
                      - icmp
                    action: reject

    State SLS example:

    .. code-block:: jinja

        {%- set filter_name = 'block-icmp' -%}
        {%- set term_name = 'first-term' -%}
        {%- set my_term_cfg = salt.netacl.get_term_pillar(filter_name, term_name) -%}

        update_icmp_first_term:
          netacl.term:
            - filter_name: {{ filter_name }}
            - filter_options:
                - not-interface-specific
            - term_name: {{ term_name }}
            - {{ my_term_cfg | json }}

    Or directly referencing the pillar keys:

    .. code-block:: yaml

        update_icmp_first_term:
          netacl.term:
            - filter_name: block-icmp
            - filter_options:
                - not-interface-specific
            - term_name: first-term
            - merge_pillar: true

    .. note::
        The first method allows the user to eventually apply complex manipulation
        and / or retrieve the data from external services before passing the
        data to the state. The second one is more straightforward, for less
        complex cases when loading the data directly from the pillar is sufficient.

    .. note::
        When passing retrieved pillar data into the state file, it is strongly
        recommended to use the json serializer explicitly (`` | json``),
        instead of relying on the default Python serializer.
    """
    ret = salt.utils.napalm.default_ret(name)
    test = __opts__["test"] or test
    if not filter_options:
        filter_options = []
    loaded = __salt__["netacl.load_term_config"](
        filter_name,
        term_name,
        filter_options=filter_options,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
        revision_id=revision_id if revision_id else name,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
        source_service=source_service,
        destination_service=destination_service,
        test=test,
        commit=commit,
        debug=debug,
        **term_fields
    )
    return salt.utils.napalm.loaded_ret(ret, loaded, test, debug)


def filter(
    name,  # pylint: disable=redefined-builtin
    filter_name,
    filter_options=None,
    terms=None,
    prepend=True,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=False,
    only_lower_merge=False,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
    test=False,
    commit=True,
    debug=False,
):
    """
    Generate and load the configuration of a policy filter.

    filter_name
        The name of the policy filter.

    filter_options
        Additional filter options. These options are platform-specific.
        See the complete list of options_.

        .. _options: https://github.com/google/capirca/wiki/Policy-format#header-section

    terms
        Dictionary of terms for this policy filter.
        If not specified or empty, will try to load the configuration from the pillar,
        unless ``merge_pillar`` is set as ``False``.

    prepend: ``True``
        When ``merge_pillar`` is set as ``True``, the final list of terms generated by merging
        the terms from ``terms`` with those defined in the pillar (if any): new terms are prepended
        at the beginning, while existing ones will preserve the position. To add the new terms
        at the end of the list, set this argument to ``False``.

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``False``
        Merge ``terms`` with the corresponding value from the pillar. Default: ``False``.

        .. note::
            By default this state does not merge, to avoid any unexpected behaviours.

            The merge logic depends on the ``prepend`` argument.

            The terms specified through the ``terms`` argument have higher priority
            than the pillar.

    only_lower_merge: ``False``
        Specify if it should merge only the terms fields. Otherwise it will try
        to merge also filters fields. Default: ``False``.
        This option requires ``merge_pillar``, otherwise it is ignored.

    revision_id
        Add a comment in the filter config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the filter configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return the changes.
        Default: ``False`` and will commit the changes on the device.

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    CLI Example:

    .. code-block:: bash

        salt 'edge01.flw01' state.sls router.acl test=True

    Output Example:

    .. code-block:: text

        edge01.flw01:
        ----------
                  ID: my-filter
            Function: netacl.filter
              Result: None
             Comment: Testing mode: Configuration discarded.
             Started: 12:24:40.598232
            Duration: 2437.139 ms
             Changes:
                      ----------
                      diff:
                          ---
                          +++
                          @@ -1228,9 +1228,24 @@
                          !
                          +ipv4 access-list my-filter
                          + 10 remark $Id: my-filter_state $
                          + 20 remark $Revision: 5 $
                          + 30 remark my-other-term
                          + 40 permit tcp any range 5678 5680 any
                          +!
                          +!
                      loaded:
                          ! $Id: my-filter_state $
                          ! $Revision: 5 $
                          no ipv6 access-list my-filter
                          ipv6 access-list my-filter
                           remark $Id: my-filter_state $
                           remark $Revision: 5 $
                           remark my-other-term
                           permit tcp any range 5678 5680 any
                          exit

        Summary for edge01.flw01
        ------------
        Succeeded: 1 (unchanged=1, changed=1)
        Failed:    0
        ------------
        Total states run:     1
        Total run time:   2.437 s

    Pillar example:

    .. code-block:: yaml

        acl:
          - my-filter:
              options:
                - inet6
              terms:
                - my-term:
                    source_port: [1234, 1235]
                    protocol:
                      - tcp
                      - udp
                    source_address: 1.2.3.4
                    action: reject
                - my-other-term:
                    source_port:
                      - [5678, 5680]
                    protocol: tcp
                    action: accept

    State SLS Example:

    .. code-block:: jinja

        {%- set filter_name = 'my-filter' -%}
        {%- set my_filter_cfg = salt.netacl.get_filter_pillar(filter_name, pillar_key='firewall') -%}
        my_first_filter_state:
          netacl.filter:
            - filter_name: {{ filter_name }}
            - options: {{ my_filter_cfg['options'] | json }}
            - terms: {{ my_filter_cfg['terms'] | json }}
            - revision_date: false
            - revision_no: 5
            - debug: true

    Or:

    .. code-block:: yaml

        my_first_filter_state:
          netacl.filter:
            - filter_name: my-filter
            - merge_pillar: true
            - pillar_key: firewall
            - revision_date: false
            - revision_no: 5
            - debug: true

    In the example above, as ``inet6`` has been specified in the ``filter_options``,
    the configuration chunk referring to ``my-term`` has been ignored as it referred to
    IPv4 only (from ``source_address`` field).

    .. note::
        The first method allows the user to eventually apply complex manipulation
        and / or retrieve the data from external services before passing the
        data to the state. The second one is more straightforward, for less
        complex cases when loading the data directly from the pillar is sufficient.

    .. note::
        When passing retrieved pillar data into the state file, it is strongly
        recommended to use the json serializer explicitly (`` | json``),
        instead of relying on the default Python serializer.
    """
    ret = salt.utils.napalm.default_ret(name)
    test = __opts__["test"] or test
    if not filter_options:
        filter_options = []
    if not terms:
        terms = []
    loaded = __salt__["netacl.load_filter_config"](
        filter_name,
        filter_options=filter_options,
        terms=terms,
        prepend=prepend,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
        only_lower_merge=only_lower_merge,
        revision_id=revision_id if revision_id else name,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
        test=test,
        commit=commit,
        debug=debug,
    )
    return salt.utils.napalm.loaded_ret(ret, loaded, test, debug)


def managed(
    name,
    filters=None,
    prepend=True,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=False,
    only_lower_merge=False,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
    test=False,
    commit=True,
    debug=False,
):
    """
    Manage the whole firewall configuration.

    filters
        Dictionary of filters for this policy.
        If not specified or empty, will try to load the configuration from the pillar,
        unless ``merge_pillar`` is set as ``False``.

    prepend: ``True``
        When ``merge_pillar`` is set as ``True``, the final list of filters generated by merging
        the filters from ``filters`` with those defined in the pillar (if any): new filters are prepended
        at the beginning, while existing ones will preserve the position. To add the new filters
        at the end of the list, set this argument to ``False``.

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``False``
        Merge the ``filters`` will the corresponding values from the pillar. Default: ``False``.

        .. note::
            By default this state does not merge, to avoid any unexpected behaviours.

            The merge logic depends on the ``prepend`` argument.

            The filters specified through the ``filters`` argument have higher priority
            than the pillar.

    only_lower_merge: ``False``
        Specify if it should merge only the filters and terms fields. Otherwise it will try
        to merge everything at the policy level. Default: ``False``.
        This option requires ``merge_pillar``, otherwise it is ignored.

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return the changes.
        Default: ``False`` and will commit the changes on the device.

    revision_id
        Add a comment in the policy config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the policy configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    CLI Example:

    .. code-block:: bash

        salt 'edge01.bjm01' state.sls router.acl test=True

    Output Example:

    .. code-block:: text

        edge01.bjm01:
        -------------
                      ID: netacl_example
                Function: netacl.managed
                  Result: None
                 Comment: Testing mode: Configuration discarded.
                 Started: 12:03:24.807023
                Duration: 5569.453 ms
                 Changes:
                          ----------
                          diff:
                              [edit firewall]
                              +    family inet {
                              +        /*
                              +         ** $Id: netacl_example $
                              +         ** $Date: 2017/07/03 $
                              +         ** $Revision: 2 $
                              +         **
                              +         */
                              +        filter my-filter {
                              +            interface-specific;
                              +            term my-term {
                              +                from {
                              +                    source-address {
                              +                        1.2.3.4/32;
                              +                    }
                              +                    protocol [ tcp udp ];
                              +                    source-port [ 1234 1235 ];
                              +                }
                              +                then {
                              +                    reject;
                              +                }
                              +            }
                              +            term my-other-term {
                              +                from {
                              +                    protocol tcp;
                              +                    source-port 5678-5680;
                              +                }
                              +                then accept;
                              +            }
                              +        }
                              +        /*
                              +         ** $Id: netacl_example $
                              +         ** $Date: 2017/07/03 $
                              +         ** $Revision: 2 $
                              +         **
                              +         */
                              +        filter block-icmp {
                              +            interface-specific;
                              +            term first-term {
                              +                from {
                              +                    protocol icmp;
                              +                }
                              +                then {
                              +                    reject;
                              +                }
                              +            }
                              +        }
                              +    }
                          loaded:
                              firewall {
                                  family inet {
                                      replace:
                                      /*
                                      ** $Id: netacl_example $
                                      ** $Date: 2017/07/03 $
                                      ** $Revision: 2 $
                                      **
                                      */
                                      filter my-filter {
                                          interface-specific;
                                          term my-term {
                                              from {
                                                  source-address {
                                                      1.2.3.4/32;
                                                  }
                                                  protocol [ tcp udp ];
                                                  source-port [ 1234 1235 ];
                                              }
                                              then {
                                                  reject;
                                              }
                                          }
                                          term my-other-term {
                                              from {
                                                  protocol tcp;
                                                  source-port 5678-5680;
                                              }
                                              then accept;
                                          }
                                      }
                                  }
                              }
                              firewall {
                                  family inet {
                                      replace:
                                      /*
                                      ** $Id: netacl_example $
                                      ** $Date: 2017/07/03 $
                                      ** $Revision: 2 $
                                      **
                                      */
                                      filter block-icmp {
                                          interface-specific;
                                          term first-term {
                                              from {
                                                  protocol icmp;
                                              }
                                              then {
                                                  reject;
                                              }
                                          }
                                      }
                                  }
                              }

            Summary for edge01.bjm01
            ------------
            Succeeded: 1 (unchanged=1, changed=1)
            Failed:    0
            ------------
            Total states run:     1
            Total run time:   5.569 s

    The policy configuration has been loaded from the pillar, having the following structure:

    .. code-block:: yaml

        firewall:
          - my-filter:
              terms:
                - my-term:
                    source_port: [1234, 1235]
                    protocol:
                      - tcp
                      - udp
                    source_address: 1.2.3.4
                    action: reject
                - my-other-term:
                    source_port:
                      - [5678, 5680]
                    protocol: tcp
                    action: accept
          - block-icmp:
              terms:
                - first-term:
                    protocol:
                      - icmp
                    action: reject

    Example SLS file:

    .. code-block:: jinja

        {%- set fw_filters = pillar.get('firewall', {}) -%}
        netacl_example:
          netacl.managed:
            - filters: {{ fw_filters | json }}
            - revision_no: 2
            - debug: true

    Or:

    .. code-block:: yaml

        netacl_example:
          netacl.managed:
            - pillar_key: firewall
            - merge_pillar: true
            - revision_no: 2
            - debug: true

    .. note::
        The first method allows the user to eventually apply complex manipulation
        and / or retrieve the data from external services before passing the
        data to the state. The second one is more straightforward, for less
        complex cases when loading the data directly from the pillar is sufficient.

    .. note::
        When passing retrieved pillar data into the state file, it is strongly
        recommended to use the json serializer explicitly (`` | json``),
        instead of relying on the default Python serializer.
    """
    ret = salt.utils.napalm.default_ret(name)
    test = __opts__["test"] or test
    if not filters:
        filters = []
    loaded = __salt__["netacl.load_policy_config"](
        filters=filters,
        prepend=prepend,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
        only_lower_merge=only_lower_merge,
        revision_id=revision_id if revision_id else name,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
        test=test,
        commit=commit,
        debug=debug,
    )
    return salt.utils.napalm.loaded_ret(ret, loaded, test, debug)



"""
    :codeauthor: :email: `Mike Place <mp@saltstack.com>`
"""


import io

from salt import template
from tests.support.mock import MagicMock
from tests.support.unit import TestCase


class TemplateTestCase(TestCase):

    render_dict = {
        "jinja": "fake_jinja_func",
        "json": "fake_json_func",
        "mako": "fake_make_func",
    }

    def test_compile_template_bad_type(self):
        """
        Test to ensure that unsupported types cannot be passed to the template compiler
        """
        ret = template.compile_template(["1", "2", "3"], None, None, None, None)
        self.assertDictEqual(ret, {})

    def test_compile_template_preserves_windows_newlines(self):
        """
        Test to ensure that a file with Windows newlines, when rendered by a
        template renderer, does not eat the CR character.
        """

        def _get_rend(renderer, value):
            """
            We need a new MagicMock each time since we're dealing with StringIO
            objects which are read like files.
            """
            return {renderer: MagicMock(return_value=io.StringIO(value))}

        input_data_windows = "foo\r\nbar\r\nbaz\r\n"
        input_data_non_windows = input_data_windows.replace("\r\n", "\n")
        renderer = "test"
        blacklist = whitelist = []

        ret = template.compile_template(
            ":string:",
            _get_rend(renderer, input_data_non_windows),
            renderer,
            blacklist,
            whitelist,
            input_data=input_data_windows,
        ).read()
        # Even though the mocked renderer returned a string without the windows
        # newlines, the compiled template should still have them.
        self.assertEqual(ret, input_data_windows)

        # Now test that we aren't adding them in unnecessarily.
        ret = template.compile_template(
            ":string:",
            _get_rend(renderer, input_data_non_windows),
            renderer,
            blacklist,
            whitelist,
            input_data=input_data_non_windows,
        ).read()
        self.assertEqual(ret, input_data_non_windows)

        # Finally, ensure that we're not unnecessarily replacing the \n with
        # \r\n in the event that the renderer returned a string with the
        # windows newlines intact.
        ret = template.compile_template(
            ":string:",
            _get_rend(renderer, input_data_windows),
            renderer,
            blacklist,
            whitelist,
            input_data=input_data_windows,
        ).read()
        self.assertEqual(ret, input_data_windows)

    def test_check_render_pipe_str(self):
        """
        Check that all renderers specified in the pipe string are available.
        """
        ret = template.check_render_pipe_str("jinja|json", self.render_dict, None, None)
        self.assertIn(("fake_jinja_func", ""), ret)
        self.assertIn(("fake_json_func", ""), ret)
        self.assertNotIn(("OBVIOUSLY_NOT_HERE", ""), ret)

    def test_check_renderer_blacklisting(self):
        """
        Check that all renderers specified in the pipe string are available.
        """
        ret = template.check_render_pipe_str(
            "jinja|json", self.render_dict, ["jinja"], None
        )
        self.assertListEqual([("fake_json_func", "")], ret)
        ret = template.check_render_pipe_str(
            "jinja|json", self.render_dict, None, ["jinja"]
        )
        self.assertListEqual([("fake_jinja_func", "")], ret)
        ret = template.check_render_pipe_str(
            "jinja|json", self.render_dict, ["jinja"], ["jinja"]
        )
        self.assertListEqual([], ret)
        ret = template.check_render_pipe_str(
            "jinja|json", self.render_dict, ["jinja"], ["jinja", "json"]
        )
        self.assertListEqual([("fake_json_func", "")], ret)



import pytest
import salt.renderers.nacl as nacl
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {nacl: {}}


def test__decrypt_object():
    """
    test _decrypt_object
    """
    secret = "Use more salt."
    crypted = "NACL[MRN3cc+fmdxyQbz6WMF+jq1hKdU5X5BBI7OjK+atvHo1ll+w1gZ7XyWtZVfq9gK9rQaMfkDxmidJKwE0Mw==]"

    secret_map = {"secret": secret}
    crypted_map = {"secret": crypted}

    secret_list = [secret]
    crypted_list = [crypted]

    with patch.dict(nacl.__salt__, {"nacl.dec": MagicMock(return_value=secret)}):
        assert nacl._decrypt_object(secret) == secret
        assert nacl._decrypt_object(crypted) == secret
        assert nacl._decrypt_object(crypted_map) == secret_map
        assert nacl._decrypt_object(crypted_list) == secret_list
        assert nacl._decrypt_object(None) is None


def test_render():
    """
    test render
    """
    secret = "Use more salt."
    crypted = "NACL[MRN3cc+fmdxyQbz6WMF+jq1hKdU5X5BBI7OjK+atvHo1ll+w1gZ7XyWtZVfq9gK9rQaMfkDxmidJKwE0Mw==]"
    with patch.dict(nacl.__salt__, {"nacl.dec": MagicMock(return_value=secret)}):
        assert nacl.render(crypted) == secret



"""
unit tests for the libvirt_events engine
"""

import salt.engines.libvirt_events as libvirt_events
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class EngineLibvirtEventTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.engine.libvirt_events
    """

    def setup_loader_modules(self):
        patcher = patch("salt.engines.libvirt_events.libvirt")
        self.mock_libvirt = patcher.start()
        self.mock_libvirt.getVersion.return_value = 2000000
        self.mock_libvirt.virEventRunDefaultImpl.return_value = (
            -1
        )  # Don't loop for ever
        self.mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE = 0
        self.mock_libvirt.VIR_DOMAIN_EVENT_ID_REBOOT = 1
        self.mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE = 0
        self.mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_REFRESH = 1
        self.mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE = 0
        self.mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_UPDATE = 1
        self.addCleanup(patcher.stop)
        self.addCleanup(delattr, self, "mock_libvirt")
        return {libvirt_events: {}}

    @patch(
        "salt.engines.libvirt_events.libvirt",
        VIR_PREFIX_NONE=0,
        VIR_PREFIX_ONE=1,
        VIR_PREFIX_TWO=2,
        VIR_PREFIX_SUB_FOO=0,
        VIR_PREFIX_SUB_BAR=1,
        VIR_PREFIX_SUB_FOOBAR=2,
    )
    def test_get_libvirt_enum_string_subprefix(self, libvirt_mock):
        """
        Make sure the libvirt enum value to string works reliably with
        elements with a sub prefix, eg VIR_PREFIX_SUB_* in this case.
        """
        # Test case with a sub prefix

        assert libvirt_events._get_libvirt_enum_string("VIR_PREFIX_", 2) == "two"

    @patch(
        "salt.engines.libvirt_events.libvirt", VIR_PREFIX_FOO=0, VIR_PREFIX_BAR_FOO=1
    )
    def test_get_libvirt_enum_string_underscores(self, libvirt_mock):
        """
        Make sure the libvirt enum value to string works reliably and items
        with an underscore aren't confused with sub prefixes.
        """
        assert libvirt_events._get_libvirt_enum_string("VIR_PREFIX_", 1) == "bar foo"

    @patch(
        "salt.engines.libvirt_events.libvirt",
        VIR_DOMAIN_EVENT_CRASHED_PANICKED=0,
        VIR_DOMAIN_EVENT_DEFINED=0,
        VIR_DOMAIN_EVENT_UNDEFINED=1,
        VIR_DOMAIN_EVENT_CRASHED=2,
        VIR_DOMAIN_EVENT_DEFINED_ADDED=0,
        VIR_DOMAIN_EVENT_DEFINED_UPDATED=1,
    )
    def test_get_domain_event_detail(self, mock_libvirt):
        """
        Test get_domain_event_detail function
        """
        assert libvirt_events._get_domain_event_detail(1, 2) == ("undefined", "unknown")
        assert libvirt_events._get_domain_event_detail(0, 1) == ("defined", "updated")
        assert libvirt_events._get_domain_event_detail(4, 2) == ("unknown", "unknown")

    @patch("salt.engines.libvirt_events.libvirt", VIR_NETWORK_EVENT_ID_LIFECYCLE=1000)
    def test_event_register(self, mock_libvirt):
        """
        Test that the libvirt_events engine actually registers events catch them and cleans
        before leaving the place.
        """
        mock_cnx = MagicMock()
        mock_libvirt.openReadOnly.return_value = mock_cnx

        # Don't loop for ever
        mock_libvirt.virEventRunDefaultImpl.return_value = -1

        mock_cnx.networkEventRegisterAny.return_value = 10000

        libvirt_events.start("test:///", "test/prefix")

        # Check that the connection has been opened
        mock_libvirt.openReadOnly.assert_called_once_with("test:///")

        # Check that the connection has been closed
        mock_cnx.close.assert_called_once()

        # Check events registration and deregistration
        mock_cnx.domainEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
            libvirt_events._domain_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
        )
        mock_cnx.networkEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_NETWORK_EVENT_ID_LIFECYCLE,
            libvirt_events._network_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "network", "event": "lifecycle"},
        )
        mock_cnx.storagePoolEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_LIFECYCLE,
            libvirt_events._pool_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "pool", "event": "lifecycle"},
        )
        mock_cnx.storagePoolEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_STORAGE_POOL_EVENT_ID_REFRESH,
            libvirt_events._pool_event_refresh_cb,
            {"prefix": "test/prefix", "object": "pool", "event": "refresh"},
        )
        mock_cnx.nodeDeviceEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_LIFECYCLE,
            libvirt_events._nodedev_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "nodedev", "event": "lifecycle"},
        )
        mock_cnx.nodeDeviceEventRegisterAny.assert_any_call(
            None,
            mock_libvirt.VIR_NODE_DEVICE_EVENT_ID_UPDATE,
            libvirt_events._nodedev_event_update_cb,
            {"prefix": "test/prefix", "object": "nodedev", "event": "update"},
        )

        # Check that the deregister events are called with the result of register
        mock_cnx.networkEventDeregisterAny.assert_called_with(
            mock_cnx.networkEventRegisterAny.return_value
        )

        # Check that the default 'all' filter actually worked
        counts = {
            obj: len(callback_def)
            for obj, callback_def in libvirt_events.CALLBACK_DEFS.items()
        }
        for obj, count in counts.items():
            register = libvirt_events.REGISTER_FUNCTIONS[obj]
            assert getattr(mock_cnx, register).call_count == count

    def test_event_skipped(self):
        """
        Test that events are skipped if their ID isn't defined in the libvirt
        module (older libvirt)
        """
        self.mock_libvirt.mock_add_spec(
            [
                "openReadOnly",
                "virEventRegisterDefaultImpl",
                "virEventRunDefaultImpl",
                "VIR_DOMAIN_EVENT_ID_LIFECYCLE",
            ],
            spec_set=True,
        )

        libvirt_events.start("test:///", "test/prefix")

        # Check events registration and deregistration
        mock_cnx = self.mock_libvirt.openReadOnly.return_value

        mock_cnx.domainEventRegisterAny.assert_any_call(
            None,
            self.mock_libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
            libvirt_events._domain_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
        )

        # Network events should have been skipped
        mock_cnx.networkEventRegisterAny.assert_not_called()

    def test_event_filtered(self):
        """
        Test that events are skipped if their ID isn't defined in the libvirt
        module (older libvirt)
        """
        libvirt_events.start("test", "test/prefix", "domain/lifecycle")

        # Check events registration and deregistration
        mock_cnx = self.mock_libvirt.openReadOnly.return_value

        mock_cnx.domainEventRegisterAny.assert_any_call(
            None,
            0,
            libvirt_events._domain_event_lifecycle_cb,
            {"prefix": "test/prefix", "object": "domain", "event": "lifecycle"},
        )

        # Network events should have been filtered out
        mock_cnx.networkEventRegisterAny.assert_not_called()



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""


import salt.modules.composer as composer
from salt.exceptions import (
    CommandExecutionError,
    CommandNotFoundError,
    SaltInvocationError,
)
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ComposerTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.composer
    """

    def setup_loader_modules(self):
        return {composer: {}}

    def test_install(self):
        """
        Test for Install composer dependencies for a directory.
        """

        # Test _valid_composer=False throws exception
        mock = MagicMock(return_value=False)
        with patch.object(composer, "_valid_composer", mock):
            self.assertRaises(CommandNotFoundError, composer.install, "d")

        # Test no directory specified throws exception
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            self.assertRaises(SaltInvocationError, composer.install, None)

        # Test `composer install` exit status != 0 throws exception
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value={"retcode": 1, "stderr": "A"})
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(CommandExecutionError, composer.install, "d")

        # Test success with quiet=True returns True
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value={"retcode": 0, "stderr": "A"})
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertTrue(
                    composer.install(
                        "dir",
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        True,
                    )
                )

        # Test success with quiet=False returns object
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            rval = {"retcode": 0, "stderr": "A", "stdout": "B"}
            mock = MagicMock(return_value=rval)
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertEqual(composer.install("dir"), rval)

    def test_update(self):
        """
        Test for Update composer dependencies for a directory.
        """

        # Test _valid_composer=False throws exception
        mock = MagicMock(return_value=False)
        with patch.object(composer, "_valid_composer", mock):
            self.assertRaises(CommandNotFoundError, composer.update, "d")

        # Test no directory specified throws exception
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=True)
            with patch.object(composer, "did_composer_install", mock):
                self.assertRaises(SaltInvocationError, composer.update, None)

        # Test update with error exit status throws exception
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=True)
            with patch.object(composer, "did_composer_install", mock):
                mock = MagicMock(return_value={"retcode": 1, "stderr": "A"})
                with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                    self.assertRaises(CommandExecutionError, composer.update, "d")

        # Test update with existing vendor directory and quiet=True
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=True)
            with patch.object(composer, "did_composer_install", mock):
                mock = MagicMock(return_value={"retcode": 0, "stderr": "A"})
                with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                    self.assertTrue(
                        composer.update(
                            "dir",
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            True,
                        )
                    )

        # Test update with no vendor directory and quiet=True
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=False)
            with patch.object(composer, "did_composer_install", mock):
                mock = MagicMock(return_value={"retcode": 0, "stderr": "A"})
                with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                    self.assertTrue(
                        composer.update(
                            "dir",
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            None,
                            True,
                        )
                    )

        # Test update with existing vendor directory
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=True)
            with patch.object(composer, "did_composer_install", mock):
                rval = {"retcode": 0, "stderr": "A", "stdout": "B"}
                mock = MagicMock(return_value=rval)
                with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                    self.assertEqual(composer.update("dir"), rval)

        # Test update with no vendor directory
        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value=False)
            with patch.object(composer, "did_composer_install", mock):
                rval = {"retcode": 0, "stderr": "A", "stdout": "B"}
                mock = MagicMock(return_value=rval)
                with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                    self.assertEqual(composer.update("dir"), rval)

    def test_selfupdate(self):
        """
        Test for Composer selfupdate
        """
        mock = MagicMock(return_value=False)
        with patch.object(composer, "_valid_composer", mock):
            self.assertRaises(CommandNotFoundError, composer.selfupdate)

        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value={"retcode": 1, "stderr": "A"})
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(CommandExecutionError, composer.selfupdate)

        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            mock = MagicMock(return_value={"retcode": 0, "stderr": "A"})
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertTrue(composer.selfupdate(quiet=True))

        mock = MagicMock(return_value=True)
        with patch.object(composer, "_valid_composer", mock):
            rval = {"retcode": 0, "stderr": "A", "stdout": "B"}
            mock = MagicMock(return_value=rval)
            with patch.dict(composer.__salt__, {"cmd.run_all": mock}):
                self.assertEqual(composer.selfupdate(), rval)



"""
tests for pkgrepo states
"""

import os

import pytest
import salt.utils.files
import salt.utils.platform
from saltfactories.utils.tempfiles import temp_file
from tests.support.case import ModuleCase
from tests.support.helpers import requires_system_grains, runs_on
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf


@skipIf(salt.utils.platform.is_windows(), "minion is windows")
@pytest.mark.destructive_test
class PkgrepoTest(ModuleCase, SaltReturnAssertsMixin):
    """
    pkgrepo state tests
    """

    @pytest.mark.requires_salt_states("pkgrepo.managed")
    @requires_system_grains
    def test_pkgrepo_01_managed(self, grains):
        """
        Test adding a repo
        """

        if grains["os_family"] == "Debian":
            try:
                from aptsources import sourceslist  # pylint: disable=unused-import
            except ImportError:
                self.skipTest("aptsources.sourceslist python module not found")
        ret = self.run_function("state.sls", mods="pkgrepo.managed", timeout=120)
        # If the below assert fails then no states were run, and the SLS in
        # tests/integration/files/file/base/pkgrepo/managed.sls needs to be
        # corrected.
        self.assertReturnNonEmptySaltType(ret)
        for state_id, state_result in ret.items():
            self.assertSaltTrueReturn(dict([(state_id, state_result)]))

    @pytest.mark.requires_salt_states("pkgrepo.absent")
    @requires_system_grains
    def test_pkgrepo_02_absent(self, grains):
        """
        Test removing the repo from the above test
        """

        ret = self.run_function("state.sls", mods="pkgrepo.absent", timeout=120)
        # If the below assert fails then no states were run, and the SLS in
        # tests/integration/files/file/base/pkgrepo/absent.sls needs to be
        # corrected.
        self.assertReturnNonEmptySaltType(ret)
        for state_id, state_result in ret.items():
            self.assertSaltTrueReturn(dict([(state_id, state_result)]))

    @pytest.mark.requires_salt_states("pkgrepo.absent", "pkgrepo.managed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkgrepo_03_with_comments(self, grains):
        """
        Test adding a repo with comments
        """
        kwargs = {}
        if grains["os_family"] == "RedHat":
            kwargs = {
                "name": "examplerepo",
                "baseurl": "http://example.com/repo",
                "enabled": False,
                "comments": ["This is a comment"],
            }
        else:
            self.skipTest(
                "{}/{} test case needed".format(grains["os_family"], grains["os"])
            )

        try:
            # Run the state to add the repo
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)

            # Run again with modified comments
            kwargs["comments"].append("This is another comment")
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)
            ret = ret[next(iter(ret))]
            self.assertEqual(
                ret["changes"],
                {
                    "comments": {
                        "old": ["This is a comment"],
                        "new": ["This is a comment", "This is another comment"],
                    }
                },
            )

            # Run a third time, no changes should be made
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)
            ret = ret[next(iter(ret))]
            self.assertFalse(ret["changes"])
            self.assertEqual(
                ret["comment"],
                "Package repo '{}' already configured".format(kwargs["name"]),
            )
        finally:
            # Clean up
            self.run_state("pkgrepo.absent", name=kwargs["name"])

    @pytest.mark.requires_salt_states("pkgrepo.managed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkgrepo_04_apt_with_architectures(self, grains):
        """
        Test managing a repo with architectures specified
        """
        if grains["os_family"].lower() != "debian":
            self.skipTest("APT-only test")

        name = "deb {{arch}}http://foo.com/bar/latest {oscodename} main".format(
            oscodename=grains["oscodename"]
        )

        def _get_arch(arch):
            return "[arch={}] ".format(arch) if arch else ""

        def _run(arch="", test=False):
            ret = self.run_state(
                "pkgrepo.managed",
                name=name.format(arch=_get_arch(arch)),
                file=fn_,
                refresh=False,
                test=test,
            )
            return ret[next(iter(ret))]

        fn_ = salt.utils.files.mkstemp(dir="/etc/apt/sources.list.d", suffix=".list")

        try:
            # Run with test=True
            ret = _run(test=True)
            assert ret["changes"] == {"repo": name.format(arch="")}, ret["changes"]
            assert "would be" in ret["comment"], ret["comment"]
            assert ret["result"] is None, ret["result"]

            # Run for real
            ret = _run()
            assert ret["changes"] == {"repo": name.format(arch="")}, ret["changes"]
            assert ret["comment"].startswith("Configured"), ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run again with test=True, should exit with no changes and a True
            # result.
            ret = _run(test=True)
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run for real again, results should be the same as above (i.e. we
            # should never get to the point where we exit with a None result).
            ret = _run()
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]

            expected_changes = {
                "line": {
                    "new": name.format(arch=_get_arch("amd64")),
                    "old": name.format(arch=""),
                },
                "architectures": {"new": ["amd64"], "old": []},
            }

            # Run with test=True and the architecture set. We should get a None
            # result with some expected changes.
            ret = _run(arch="amd64", test=True)
            assert ret["changes"] == expected_changes, ret["changes"]
            assert "would be" in ret["comment"], ret["comment"]
            assert ret["result"] is None, ret["result"]

            # Run for real, with the architecture set. We should get a True
            # result with the same changes.
            ret = _run(arch="amd64")
            assert ret["changes"] == expected_changes, ret["changes"]
            assert ret["comment"].startswith("Configured"), ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run again with test=True, should exit with no changes and a True
            # result.
            ret = _run(arch="amd64", test=True)
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run for real again, results should be the same as above (i.e. we
            # should never get to the point where we exit with a None result).
            ret = _run(arch="amd64")
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]

            expected_changes = {
                "line": {
                    "new": name.format(arch=""),
                    "old": name.format(arch=_get_arch("amd64")),
                },
                "architectures": {"new": [], "old": ["amd64"]},
            }

            # Run with test=True and the architecture set back to the original
            # value. We should get a None result with some expected changes.
            ret = _run(test=True)
            assert ret["changes"] == expected_changes, ret["changes"]
            assert "would be" in ret["comment"], ret["comment"]
            assert ret["result"] is None, ret["result"]

            # Run for real, with the architecture set. We should get a True
            # result with the same changes.
            ret = _run()
            assert ret["changes"] == expected_changes, ret["changes"]
            assert ret["comment"].startswith("Configured"), ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run again with test=True, should exit with no changes and a True
            # result.
            ret = _run(test=True)
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]

            # Run for real again, results should be the same as above (i.e. we
            # should never get to the point where we exit with a None result).
            ret = _run()
            assert not ret["changes"], ret["changes"]
            assert "already" in ret["comment"], ret["comment"]
            assert ret["result"] is True, ret["result"]
        finally:
            try:
                os.remove(fn_)
            except OSError:
                pass

    @pytest.mark.requires_salt_states("pkgrepo.absent", "pkgrepo.managed")
    @requires_system_grains
    @pytest.mark.slow_test
    def test_pkgrepo_05_copr_with_comments(self, grains):
        """
        Test copr
        """
        kwargs = {}
        if grains["os_family"] == "RedHat":
            if (
                grains["osfinger"] == "CentOS Linux-7"
                or grains["osfinger"] == "Amazon Linux-2"
            ):
                self.skipTest("copr plugin not installed on Centos 7 CI")
            kwargs = {
                "name": "hello-copr",
                "copr": "mymindstorm/hello",
                "enabled": False,
                "comments": ["This is a comment"],
            }
        else:
            self.skipTest(
                "{}/{} test case needed".format(grains["os_family"], grains["os"])
            )

        try:
            # Run the state to add the repo
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)

            # Run again with modified comments
            kwargs["comments"].append("This is another comment")
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)
            ret = ret[next(iter(ret))]
            self.assertEqual(
                ret["changes"],
                {
                    "comments": {
                        "old": ["This is a comment"],
                        "new": ["This is a comment", "This is another comment"],
                    }
                },
            )

            # Run a third time, no changes should be made
            ret = self.run_state("pkgrepo.managed", **kwargs)
            self.assertSaltTrueReturn(ret)
            ret = ret[next(iter(ret))]
            self.assertFalse(ret["changes"])
            self.assertEqual(
                ret["comment"],
                "Package repo '{}' already configured".format(kwargs["name"]),
            )
        finally:
            # Clean up
            self.run_state("pkgrepo.absent", copr=kwargs["copr"])

    @runs_on(kernel="linux", os="Ubuntu")
    def test_managed_multiple_comps(self):
        state_file = """
        ubuntu-backports:
          pkgrepo.managed:
            - name: 'deb http://fi.archive.ubuntu.com/ubuntu focal-backports'
            - comps: main, restricted, universe, multiverse
            - refresh: false
            - disabled: false
            - clean_file: true
            - file: /etc/apt/sources.list.d/99-salt-archive-ubuntu-focal-backports.list
            - require_in:
              - pkgrepo: canonical-ubuntu

        canonical-ubuntu:
          pkgrepo.managed:
            - name: 'deb http://archive.canonical.com/ubuntu {{ salt['grains.get']('oscodename') }}'
            - comps: partner
            - refresh: false
            - disabled: false
            - clean_file: true
            - file: /etc/apt/sources.list.d/99-salt-canonical-ubuntu.list
        """

        def remove_apt_list_file(path):
            if os.path.exists(path):
                os.unlink(path)

        self.addCleanup(
            remove_apt_list_file,
            "/etc/apt/sources.list.d/99-salt-canonical-ubuntu.list",
        )
        self.addCleanup(
            remove_apt_list_file,
            "/etc/apt/sources.list.d/99-salt-archive-ubuntu-focal-backports.list",
        )
        with temp_file(
            "multiple-comps-repos.sls", state_file, RUNTIME_VARS.TMP_BASEENV_STATE_TREE
        ):
            ret = self.run_function("state.sls", ["multiple-comps-repos"])
            for state_run in ret.values():
                # On the first run, we must have changes
                assert state_run["changes"]
            ret = self.run_function("state.sls", ["multiple-comps-repos"])
            for state_run in ret.values():
                # On the second run though, we shouldn't have changes made
                assert not state_run["changes"]



import requests


def test_server(host_ip, http_port):
    '''
    Test that wordpress was setup on the minion correctly and returns a 200 after being installed
    '''
    resp = requests.get('http://{}:{}'.format(host_ip, http_port), headers={'Host': 'blog.manfred.io'})
    assert resp.status_code == 200



"""
    :codeauthor: Dave Rawks (dave@pandora.com)


    tests.unit.modules.parted_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import salt.modules.parted_partition as parted
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class PartedTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        self.cmdrun = MagicMock()
        self.cmdrun_stdout = MagicMock()
        self.addCleanup(delattr, self, "cmdrun")
        self.addCleanup(delattr, self, "cmdrun_stdout")
        return {
            parted: {
                "__salt__": {
                    "cmd.run": self.cmdrun,
                    "cmd.run_stdout": self.cmdrun_stdout,
                }
            }
        }

    # Test __virtual__ function for module registration

    def test_virtual_bails_on_windows(self):
        """
        If running windows, __virtual__ shouldn't register module
        """
        with patch("salt.utils.platform.is_windows", lambda: True):
            ret = parted.__virtual__()
            err = (
                False,
                "The parted execution module failed to load Windows systems are not"
                " supported.",
            )
            self.assertEqual(err, ret)

    def test_virtual_bails_without_parted(self):
        """
        If parted not in PATH, __virtual__ shouldn't register module
        """
        with patch("salt.utils.path.which", lambda exe: not exe == "parted"), patch(
            "salt.utils.platform.is_windows", return_value=False
        ):
            ret = parted.__virtual__()
            err = (
                False,
                "The parted execution module failed to load parted binary is not in the"
                " path.",
            )
            self.assertEqual(err, ret)

    def test_virtual_bails_without_lsblk(self):
        """
        If lsblk not in PATH, __virtual__ shouldn't register module
        """
        with patch("salt.utils.path.which", lambda exe: not exe == "lsblk"), patch(
            "salt.utils.platform.is_windows", return_value=False
        ):
            ret = parted.__virtual__()
            err = (
                False,
                "The parted execution module failed to load lsblk binary is not in the"
                " path.",
            )
            self.assertEqual(err, ret)

    def test_virtual_bails_without_partprobe(self):
        """
        If partprobe not in PATH, __virtual__ shouldn't register module
        """
        with patch("salt.utils.path.which", lambda exe: not exe == "partprobe"), patch(
            "salt.utils.platform.is_windows", return_value=False
        ):
            ret = parted.__virtual__()
            err = (
                False,
                "The parted execution module failed to load partprobe binary is not in"
                " the path.",
            )
            self.assertEqual(err, ret)

    def test_virtual(self):
        """
        On expected platform with correct utils in PATH, register "partition" module
        """
        with patch("salt.utils.platform.is_windows", lambda: False), patch(
            "salt.utils.path.which", lambda exe: exe in ("parted", "lsblk", "partprobe")
        ):
            ret = parted.__virtual__()
            expect = "partition"
            self.assertEqual(ret, expect)

    # Test probe function

    def test_probe_wo_args(self):
        parted.probe()
        self.cmdrun.assert_called_once_with("partprobe -- ")

    def test_probe_w_single_arg(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            parted.probe("/dev/sda")
            self.cmdrun.assert_called_once_with("partprobe -- /dev/sda")

    def test_probe_w_multiple_args(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            parted.probe("/dev/sda", "/dev/sdb")
            self.cmdrun.assert_called_once_with("partprobe -- /dev/sda /dev/sdb")

    # Test _list function

    @staticmethod
    def parted_print_output(k):
        output = {
            "valid": (
                """BYT;\n"""
                """/dev/sda:4000GB:scsi:512:512:gpt:AMCC 9650SE-24M DISK:;\n"""
                """1:17.4kB:150MB:150MB:ext3::boot;\n"""
                """2:3921GB:4000GB:79.3GB:linux-swap(v1)::;\n"""
            ),
            "valid chs": (
                """CHS;\n"""
                """/dev/sda:3133,0,2:scsi:512:512:gpt:AMCC 9650SE-24M DISK:;\n"""
                """1:0,0,34:2431,134,43:ext3::boot;\n"""
                """2:2431,134,44:2492,80,42:linux-swap(v1)::;\n"""
            ),
            "valid_legacy": (
                """BYT;\n"""
                """/dev/sda:4000GB:scsi:512:512:gpt:AMCC 9650SE-24M DISK;\n"""
                """1:17.4kB:150MB:150MB:ext3::boot;\n"""
                """2:3921GB:4000GB:79.3GB:linux-swap(v1)::;\n"""
            ),
            "empty": "",
            "bad_label_info": (
                """BYT;\n"""
                """badbadbadbad\n"""
                """1:17.4kB:150MB:150MB:ext3::boot;\n"""
                """2:3921GB:4000GB:79.3GB:linux-swap(v1)::;\n"""
            ),
            "bad_header": (
                """badbadbadbad\n"""
                """/dev/sda:4000GB:scsi:512:512:gpt:AMCC 9650SE-24M DISK:;\n"""
                """1:17.4kB:150MB:150MB:ext3::boot;\n"""
                """2:3921GB:4000GB:79.3GB:linux-swap(v1)::;\n"""
            ),
            "bad_partition": (
                """BYT;\n"""
                """/dev/sda:4000GB:scsi:512:512:gpt:AMCC 9650SE-24M DISK:;\n"""
                """badbadbadbad\n"""
                """2:3921GB:4000GB:79.3GB:linux-swap(v1)::;\n"""
            ),
        }
        return output[k]

    def test_list__without_device(self):
        self.assertRaises(TypeError, parted.list_)

    def test_list__empty_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("empty")
            output = parted.list_("/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")
            expected = {"info": {}, "partitions": {}}
            self.assertEqual(output, expected)

    def test_list__valid_unit_empty_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("empty")
            output = parted.list_("/dev/sda", unit="s")
            self.cmdrun_stdout.assert_called_once_with(
                "parted -m -s /dev/sda unit s print"
            )
            expected = {"info": {}, "partitions": {}}
            self.assertEqual(output, expected)

    def test_list__invalid_unit(self):
        self.assertRaises(
            CommandExecutionError, parted.list_, "/dev/sda", unit="badbadbad"
        )
        self.assertFalse(self.cmdrun.called)

    def test_list__bad_header(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("bad_header")
            self.assertRaises(CommandExecutionError, parted.list_, "/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")

    def test_list__bad_label_info(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("bad_label_info")
            self.assertRaises(CommandExecutionError, parted.list_, "/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")

    def test_list__bad_partition(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("bad_partition")
            self.assertRaises(CommandExecutionError, parted.list_, "/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")

    def test_list__valid_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("valid")
            output = parted.list_("/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")
            expected = {
                "info": {
                    "logical sector": "512",
                    "physical sector": "512",
                    "interface": "scsi",
                    "model": "AMCC 9650SE-24M DISK",
                    "disk": "/dev/sda",
                    "disk flags": "",
                    "partition table": "gpt",
                    "size": "4000GB",
                },
                "partitions": {
                    "1": {
                        "end": "150MB",
                        "number": "1",
                        "start": "17.4kB",
                        "file system": "ext3",
                        "flags": "boot",
                        "name": "",
                        "size": "150MB",
                    },
                    "2": {
                        "end": "4000GB",
                        "number": "2",
                        "start": "3921GB",
                        "file system": "linux-swap(v1)",
                        "flags": "",
                        "name": "",
                        "size": "79.3GB",
                    },
                },
            }
            self.assertEqual(output, expected)

    def test_list__valid_unit_valid_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("valid")
            output = parted.list_("/dev/sda", unit="s")
            self.cmdrun_stdout.assert_called_once_with(
                "parted -m -s /dev/sda unit s print"
            )
            expected = {
                "info": {
                    "logical sector": "512",
                    "physical sector": "512",
                    "interface": "scsi",
                    "model": "AMCC 9650SE-24M DISK",
                    "disk": "/dev/sda",
                    "disk flags": "",
                    "partition table": "gpt",
                    "size": "4000GB",
                },
                "partitions": {
                    "1": {
                        "end": "150MB",
                        "number": "1",
                        "start": "17.4kB",
                        "file system": "ext3",
                        "flags": "boot",
                        "name": "",
                        "size": "150MB",
                    },
                    "2": {
                        "end": "4000GB",
                        "number": "2",
                        "start": "3921GB",
                        "file system": "linux-swap(v1)",
                        "flags": "",
                        "name": "",
                        "size": "79.3GB",
                    },
                },
            }
            self.assertEqual(output, expected)

    def test_list__valid_unit_chs_valid_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("valid chs")
            output = parted.list_("/dev/sda", unit="chs")
            self.cmdrun_stdout.assert_called_once_with(
                "parted -m -s /dev/sda unit chs print"
            )
            expected = {
                "info": {
                    "logical sector": "512",
                    "physical sector": "512",
                    "interface": "scsi",
                    "model": "AMCC 9650SE-24M DISK",
                    "disk": "/dev/sda",
                    "disk flags": "",
                    "partition table": "gpt",
                    "size": "3133,0,2",
                },
                "partitions": {
                    "1": {
                        "end": "2431,134,43",
                        "number": "1",
                        "start": "0,0,34",
                        "file system": "ext3",
                        "flags": "boot",
                        "name": "",
                    },
                    "2": {
                        "end": "2492,80,42",
                        "number": "2",
                        "start": "2431,134,44",
                        "file system": "linux-swap(v1)",
                        "flags": "",
                        "name": "",
                    },
                },
            }
            self.assertEqual(output, expected)

    def test_list__valid_legacy_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("valid_legacy")
            output = parted.list_("/dev/sda")
            self.cmdrun_stdout.assert_called_once_with("parted -m -s /dev/sda print")
            expected = {
                "info": {
                    "logical sector": "512",
                    "physical sector": "512",
                    "interface": "scsi",
                    "model": "AMCC 9650SE-24M DISK",
                    "disk": "/dev/sda",
                    "partition table": "gpt",
                    "size": "4000GB",
                },
                "partitions": {
                    "1": {
                        "end": "150MB",
                        "number": "1",
                        "start": "17.4kB",
                        "file system": "ext3",
                        "flags": "boot",
                        "name": "",
                        "size": "150MB",
                    },
                    "2": {
                        "end": "4000GB",
                        "number": "2",
                        "start": "3921GB",
                        "file system": "linux-swap(v1)",
                        "flags": "",
                        "name": "",
                        "size": "79.3GB",
                    },
                },
            }
            self.assertEqual(output, expected)

    def test_list__valid_unit_valid_legacy_cmd_output(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun_stdout.return_value = self.parted_print_output("valid_legacy")
            output = parted.list_("/dev/sda", unit="s")
            self.cmdrun_stdout.assert_called_once_with(
                "parted -m -s /dev/sda unit s print"
            )
            expected = {
                "info": {
                    "logical sector": "512",
                    "physical sector": "512",
                    "interface": "scsi",
                    "model": "AMCC 9650SE-24M DISK",
                    "disk": "/dev/sda",
                    "partition table": "gpt",
                    "size": "4000GB",
                },
                "partitions": {
                    "1": {
                        "end": "150MB",
                        "number": "1",
                        "start": "17.4kB",
                        "file system": "ext3",
                        "flags": "boot",
                        "name": "",
                        "size": "150MB",
                    },
                    "2": {
                        "end": "4000GB",
                        "number": "2",
                        "start": "3921GB",
                        "file system": "linux-swap(v1)",
                        "flags": "",
                        "name": "",
                        "size": "79.3GB",
                    },
                },
            }
            self.assertEqual(output, expected)

    def test_btrfs_fstypes(self):
        """Tests if we see btrfs as valid fs type"""
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            try:
                parted.mkfs("/dev/foo", "btrfs")
            except CommandExecutionError:
                self.fail("Btrfs is not in the supported fstypes")

    def test_xfs_fstypes(self):
        """Tests if we see xfs as valid fs type"""
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            try:
                parted.mkfs("/dev/foo", "xfs")
            except CommandExecutionError:
                self.fail("XFS is not in the supported fstypes")

    def test_disk_set(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun.return_value = ""
            output = parted.disk_set("/dev/sda", "pmbr_boot", "on")
            self.cmdrun.assert_called_once_with(
                ["parted", "-m", "-s", "/dev/sda", "disk_set", "pmbr_boot", "on"]
            )
            assert output == []

    def test_disk_toggle(self):
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun.return_value = ""
            output = parted.disk_toggle("/dev/sda", "pmbr_boot")
            self.cmdrun.assert_called_once_with(
                ["parted", "-m", "-s", "/dev/sda", "disk_toggle", "pmbr_boot"]
            )
            assert output == []

    def test__is_fstype(self):
        assert parted._is_fstype("fat")
        assert not parted._is_fstype("thicc")

    def test_mkpart_without_fstype(self):
        """Test if mkpart works with an empty fstype"""
        cmd = (
            "parted",
            "-m",
            "-s",
            "--",
            "/dev/nothinghere",
            "mkpart",
            "primary",
            "",
            "",
        )
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun.return_value = ""
            output = parted.mkpart("/dev/nothinghere", "primary")
            self.cmdrun.assert_called_once_with(cmd, python_shell=False)
            assert output == []

    def test_mkpartfs_to_mkpart(self):
        """Test if mkpart got all arguments from mkpartfs"""
        cmd = (
            "parted",
            "-m",
            "-s",
            "--",
            "/dev/nothinghere",
            "mkpart",
            "primary",
            "ext3",
            "1",
            "2",
        )
        with patch("salt.modules.parted_partition._validate_device", MagicMock()):
            self.cmdrun.return_value = ""
            output = parted.mkpartfs(
                "/dev/nothinghere", "primary", fs_type="ext3", start="1", end="2"
            )
            self.cmdrun.assert_called_once_with(cmd, python_shell=False)
            assert output == []



"""
    salt._logging.mixins
    ~~~~~~~~~~~~~~~~~~~~

    Logging related mix-ins
"""


import logging
import sys


class NewStyleClassMixin:
    """
    Simple new style class to make pylint shut up!
    This is required because SaltLoggingClass can't subclass object directly:

        'Cannot create a consistent method resolution order (MRO) for bases'
    """


class LoggingProfileMixin:
    """
    Simple mix-in class to add a trace method to python's logging.
    """

    def profile(self, msg, *args, **kwargs):
        self.log(getattr(logging, "PROFILE", 15), msg, *args, **kwargs)


class LoggingTraceMixin:
    """
    Simple mix-in class to add a trace method to python's logging.
    """

    def trace(self, msg, *args, **kwargs):
        self.log(getattr(logging, "TRACE", 5), msg, *args, **kwargs)


class LoggingGarbageMixin:
    """
    Simple mix-in class to add a garbage method to python's logging.
    """

    def garbage(self, msg, *args, **kwargs):
        self.log(getattr(logging, "GARBAGE", 5), msg, *args, **kwargs)


class LoggingMixinMeta(type):
    """
    This class is called whenever a new instance of ``SaltLoggingClass`` is
    created.

    What this class does is check if any of the bases have a `trace()` or a
    `garbage()` method defined, if they don't we add the respective mix-ins to
    the bases.
    """

    def __new__(mcs, name, bases, attrs):
        include_profile = include_trace = include_garbage = True
        bases = list(bases)
        if name == "SaltLoggingClass":
            for base in bases:
                if hasattr(base, "trace"):
                    include_trace = False
                if hasattr(base, "garbage"):
                    include_garbage = False
        if include_profile:
            bases.append(LoggingProfileMixin)
        if include_trace:
            bases.append(LoggingTraceMixin)
        if include_garbage:
            bases.append(LoggingGarbageMixin)
        return super().__new__(mcs, name, tuple(bases), attrs)


class ExcInfoOnLogLevelFormatMixin:
    """
    Logging handler class mixin to properly handle including exc_info on a per logging handler basis
    """

    def format(self, record):
        """
        Format the log record to include exc_info if the handler is enabled for a specific log level
        """
        formatted_record = super().format(record)
        exc_info_on_loglevel = getattr(record, "exc_info_on_loglevel", None)
        exc_info_on_loglevel_formatted = getattr(
            record, "exc_info_on_loglevel_formatted", None
        )
        if exc_info_on_loglevel is None and exc_info_on_loglevel_formatted is None:
            return formatted_record

        # If we reached this far it means the log record was created with exc_info_on_loglevel
        # If this specific handler is enabled for that record, then we should format it to
        # include the exc_info details
        if self.level > exc_info_on_loglevel:
            # This handler is not enabled for the desired exc_info_on_loglevel, don't include exc_info
            return formatted_record

        # If we reached this far it means we should include exc_info
        if (
            not record.exc_info_on_loglevel_instance
            and not exc_info_on_loglevel_formatted
        ):
            # This should actually never occur
            return formatted_record

        if record.exc_info_on_loglevel_formatted is None:
            # Let's cache the formatted exception to avoid recurring conversions and formatting calls
            if (
                self.formatter is None
            ):  # pylint: disable=access-member-before-definition
                self.formatter = logging._defaultFormatter
            record.exc_info_on_loglevel_formatted = self.formatter.formatException(
                record.exc_info_on_loglevel_instance
            )

        # Let's format the record to include exc_info just like python's logging formatted does
        if formatted_record[-1:] != "\n":
            formatted_record += "\n"

        try:
            formatted_record += record.exc_info_on_loglevel_formatted
        except UnicodeError:
            # According to the standard library logging formatter comments:
            #
            #     Sometimes filenames have non-ASCII chars, which can lead
            #     to errors when s is Unicode and record.exc_text is str
            #     See issue 8924.
            #     We also use replace for when there are multiple
            #     encodings, e.g. UTF-8 for the filesystem and latin-1
            #     for a script. See issue 13232.
            formatted_record += record.exc_info_on_loglevel_formatted.decode(
                sys.getfilesystemencoding(), "replace"
            )
        # Reset the record.exc_info_on_loglevel_instance because it might need
        # to "travel" through a multiprocessing process and it might contain
        # data which is not pickle'able
        record.exc_info_on_loglevel_instance = None
        return formatted_record



"""
Qemu Command Wrapper

The qemu system comes with powerful tools, such as qemu-img and qemu-nbd which
are used here to build up kvm images.
"""


import glob
import logging
import os
import tempfile
import time

import salt.crypt
import salt.utils.path

# Set up logging
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if qemu-img and qemu-nbd are installed
    """
    if salt.utils.path.which("qemu-nbd"):
        return "qemu_nbd"
    return (
        False,
        "The qemu_nbd execution module cannot be loaded: the qemu-nbd binary is not in"
        " the path.",
    )


def connect(image):
    """
    Activate nbd for an image file.

    CLI Example:

    .. code-block:: bash

        salt '*' qemu_nbd.connect /tmp/image.raw
    """
    if not os.path.isfile(image):
        log.warning("Could not connect image: %s does not exist", image)
        return ""

    if salt.utils.path.which("sfdisk"):
        fdisk = "sfdisk -d"
    else:
        fdisk = "fdisk -l"
    __salt__["cmd.run"]("modprobe nbd max_part=63")
    for nbd in glob.glob("/dev/nbd?"):
        if __salt__["cmd.retcode"]("{} {}".format(fdisk, nbd)):
            while True:
                # Sometimes nbd does not "take hold", loop until we can verify
                __salt__["cmd.run"](
                    "qemu-nbd -c {} {}".format(nbd, image),
                    python_shell=False,
                )
                if not __salt__["cmd.retcode"]("{} {}".format(fdisk, nbd)):
                    break
            return nbd
    log.warning("Could not connect image: %s", image)
    return ""


def mount(nbd, root=None):
    """
    Pass in the nbd connection device location, mount all partitions and return
    a dict of mount points

    CLI Example:

    .. code-block:: bash

        salt '*' qemu_nbd.mount /dev/nbd0
    """
    __salt__["cmd.run"](
        "partprobe {}".format(nbd),
        python_shell=False,
    )
    ret = {}
    if root is None:
        root = os.path.join(tempfile.gettempdir(), "nbd", os.path.basename(nbd))
    for part in glob.glob("{}p*".format(nbd)):
        m_pt = os.path.join(root, os.path.basename(part))
        time.sleep(1)
        mnt = __salt__["mount.mount"](m_pt, part, True)
        if mnt is not True:
            continue
        ret[m_pt] = part
    return ret


def init(image, root=None):
    """
    Mount the named image via qemu-nbd and return the mounted roots

    CLI Example:

    .. code-block:: bash

        salt '*' qemu_nbd.init /srv/image.qcow2
    """
    nbd = connect(image)
    if not nbd:
        return ""
    return mount(nbd, root)


def clear(mnt):
    """
    Pass in the mnt dict returned from nbd_mount to unmount and disconnect
    the image from nbd. If all of the partitions are unmounted return an
    empty dict, otherwise return a dict containing the still mounted
    partitions

    CLI Example:

    .. code-block:: bash

        salt '*' qemu_nbd.clear '{"/mnt/foo": "/dev/nbd0p1"}'
    """
    ret = {}
    nbds = set()
    for m_pt, dev in mnt.items():
        mnt_ret = __salt__["mount.umount"](m_pt)
        if mnt_ret is not True:
            ret[m_pt] = dev
        nbds.add(dev[: dev.rindex("p")])
    if ret:
        return ret
    for nbd in nbds:
        __salt__["cmd.run"]("qemu-nbd -d {}".format(nbd), python_shell=False)
    return ret



import pytest
from salt.exceptions import CommandExecutionError


@pytest.fixture(scope="module", autouse=True)
def install_npm(sminion):
    try:
        sminion.functions.state.single("pkg.installed", name="npm")
        # Just name the thing we're looking for
        sminion.functions.npm  # pylint: disable=pointless-statement
    except (CommandExecutionError, AttributeError) as e:
        pytest.skip("Unable to install npm - " + str(e))


@pytest.mark.slow_test
@pytest.mark.destructive_test
@pytest.mark.requires_network
def test_removed_installed_cycle(states, modules):
    project_version = "pm2@5.1.0"
    success = modules.npm.uninstall("pm2")
    assert success, "Unable to uninstall pm2 in prep for tests"

    ret = states.npm.installed(name=project_version)
    assert ret.result is True, "Failed to states.npm.installed {} - {}".format(
        project_version, ret.comment
    )

    ret = states.npm.removed(name=project_version)
    assert ret.result is True, "Failed to states.npm.removed {} - {}".format(
        project_version, ret.comment
    )



import pytest
import salt.pillar

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_pillar_top_compound_match(salt_master, pillar_state_tree, grains):
    """
    Test that a compound match topfile that refers to a nodegroup via N@ works
    as expected.
    """
    top_file_contents = """
    base:
      'N@mins not L@minion':
        - ng1
      'N@missing_minion':
        - ng2
    """
    ng1_pillar_contents = "pillar_from_nodegroup: True"
    ng2_pillar_contents = "pillar_from_nodegroup_with_ghost: True"
    with pytest.helpers.temp_file(
        "top.sls", top_file_contents, pillar_state_tree
    ), pytest.helpers.temp_file(
        "ng1.sls", ng1_pillar_contents, pillar_state_tree
    ), pytest.helpers.temp_file(
        "ng2.sls", ng2_pillar_contents, pillar_state_tree
    ):
        opts = salt_master.config.copy()
        opts["nodegroups"] = {
            "min": "minion",
            "sub_min": "sub_minion",
            "mins": "N@min or N@sub_min",
            "missing_minion": "L@minion,ghostminion",
        }
        pillar_obj = salt.pillar.Pillar(opts, grains, "minion", "base")
        ret = pillar_obj.compile_pillar()
        assert ret.get("pillar_from_nodegroup_with_ghost") is True
        assert ret.get("pillar_from_nodegroup") is None

        sub_pillar_obj = salt.pillar.Pillar(opts, grains, "sub_minion", "base")
        sub_ret = sub_pillar_obj.compile_pillar()
        assert sub_ret.get("pillar_from_nodegroup_with_ghost") is None
        assert sub_ret.get("pillar_from_nodegroup") is True



"""
Jinja loading utils to enable a more powerful backend for jinja templates
"""


import atexit
import logging
import os.path
import pipes
import pprint
import re
import time
import uuid
import warnings
from collections.abc import Hashable
from functools import wraps
from xml.dom import minidom
from xml.etree.ElementTree import Element, SubElement, tostring

import jinja2
import salt.fileclient
import salt.utils.data
import salt.utils.files
import salt.utils.json
import salt.utils.stringutils
import salt.utils.url
import salt.utils.yaml
from jinja2 import BaseLoader, Markup, TemplateNotFound, nodes
from jinja2.environment import TemplateModule
from jinja2.exceptions import TemplateRuntimeError
from jinja2.ext import Extension
from salt.exceptions import TemplateError
from salt.utils.decorators.jinja import jinja_filter, jinja_global, jinja_test
from salt.utils.odict import OrderedDict
from salt.utils.versions import LooseVersion

log = logging.getLogger(__name__)

__all__ = ["SaltCacheLoader", "SerializerExtension"]

GLOBAL_UUID = uuid.UUID("91633EBF-1C86-5E33-935A-28061F4B480E")
JINJA_VERSION = LooseVersion(jinja2.__version__)


class SaltCacheLoader(BaseLoader):
    """
    A special jinja Template Loader for salt.
    Requested templates are always fetched from the server
    to guarantee that the file is up to date.
    Templates are cached like regular salt states
    and only loaded once per loader instance.
    """

    _cached_pillar_client = None
    _cached_client = None

    @classmethod
    def shutdown(cls):
        for attr in ("_cached_client", "_cached_pillar_client"):
            client = getattr(cls, attr, None)
            if client is not None:
                # PillarClient and LocalClient objects do not have a destroy method
                if hasattr(client, "destroy"):
                    client.destroy()
                setattr(cls, attr, None)

    def __init__(
        self,
        opts,
        saltenv="base",
        encoding="utf-8",
        pillar_rend=False,
        _file_client=None,
    ):
        self.opts = opts
        self.saltenv = saltenv
        self.encoding = encoding
        self.pillar_rend = pillar_rend
        if self.pillar_rend:
            if saltenv not in self.opts["pillar_roots"]:
                self.searchpath = []
            else:
                self.searchpath = opts["pillar_roots"][saltenv]
        else:
            self.searchpath = [os.path.join(opts["cachedir"], "files", saltenv)]
        log.debug("Jinja search path: %s", self.searchpath)
        self.cached = []
        self._file_client = _file_client
        # Instantiate the fileclient
        self.file_client()

    def file_client(self):
        """
        Return a file client. Instantiates on first call.
        """
        # If there was no file_client passed to the class, create a cache_client
        # and use that. This avoids opening a new file_client every time this
        # class is instantiated
        if self._file_client is None:
            attr = "_cached_pillar_client" if self.pillar_rend else "_cached_client"
            cached_client = getattr(self, attr, None)
            if cached_client is None:
                cached_client = salt.fileclient.get_file_client(
                    self.opts, self.pillar_rend
                )
                setattr(SaltCacheLoader, attr, cached_client)
            self._file_client = cached_client
        return self._file_client

    def cache_file(self, template):
        """
        Cache a file from the salt master
        """
        saltpath = salt.utils.url.create(template)
        self.file_client().get_file(saltpath, "", True, self.saltenv)

    def check_cache(self, template):
        """
        Cache a file only once
        """
        if template not in self.cached:
            self.cache_file(template)
            self.cached.append(template)

    def get_source(self, environment, template):
        """
        Salt-specific loader to find imported jinja files.

        Jinja imports will be interpreted as originating from the top
        of each of the directories in the searchpath when the template
        name does not begin with './' or '../'.  When a template name
        begins with './' or '../' then the import will be relative to
        the importing file.

        """
        # FIXME: somewhere do seprataor replacement: '\\' => '/'
        _template = template
        if template.split("/", 1)[0] in ("..", "."):
            is_relative = True
        else:
            is_relative = False
        # checks for relative '..' paths that step-out of file_roots
        if is_relative:
            # Starts with a relative path indicator

            if not environment or "tpldir" not in environment.globals:
                log.warning(
                    'Relative path "%s" cannot be resolved without an environment',
                    template,
                )
                raise TemplateNotFound(template)
            base_path = environment.globals["tpldir"]
            _template = os.path.normpath("/".join((base_path, _template)))
            if _template.split("/", 1)[0] == "..":
                log.warning(
                    'Discarded template path "%s": attempts to'
                    " ascend outside of salt://",
                    template,
                )
                raise TemplateNotFound(template)

        self.check_cache(_template)

        if environment and template:
            tpldir = os.path.dirname(_template).replace("\\", "/")
            tplfile = _template
            if is_relative:
                tpldir = environment.globals.get("tpldir", tpldir)
                tplfile = template
            tpldata = {
                "tplfile": tplfile,
                "tpldir": "." if tpldir == "" else tpldir,
                "tpldot": tpldir.replace("/", "."),
            }
            environment.globals.update(tpldata)

        # pylint: disable=cell-var-from-loop
        for spath in self.searchpath:
            filepath = os.path.join(spath, _template)
            try:
                with salt.utils.files.fopen(filepath, "rb") as ifile:
                    contents = ifile.read().decode(self.encoding)
                    mtime = os.path.getmtime(filepath)

                    def uptodate():
                        try:
                            return os.path.getmtime(filepath) == mtime
                        except OSError:
                            return False

                    return contents, filepath, uptodate
            except OSError:
                # there is no file under current path
                continue
        # pylint: enable=cell-var-from-loop

        # there is no template file within searchpaths
        raise TemplateNotFound(template)


atexit.register(SaltCacheLoader.shutdown)


class PrintableDict(OrderedDict):
    """
    Ensures that dict str() and repr() are YAML friendly.

    .. code-block:: python

        mapping = OrderedDict([('a', 'b'), ('c', None)])
        print mapping
        # OrderedDict([('a', 'b'), ('c', None)])

        decorated = PrintableDict(mapping)
        print decorated
        # {'a': 'b', 'c': None}
    """

    def __str__(self):
        output = []
        for key, value in self.items():
            if isinstance(value, str):
                # keeps quotes around strings
                # pylint: disable=repr-flag-used-in-string
                output.append("{!r}: {!r}".format(key, value))
                # pylint: enable=repr-flag-used-in-string
            else:
                # let default output
                output.append("{!r}: {!s}".format(key, value))
        return "{" + ", ".join(output) + "}"

    def __repr__(self):  # pylint: disable=W0221
        output = []
        for key, value in self.items():
            # Raw string formatter required here because this is a repr
            # function.
            # pylint: disable=repr-flag-used-in-string
            output.append("{!r}: {!r}".format(key, value))
            # pylint: enable=repr-flag-used-in-string
        return "{" + ", ".join(output) + "}"


# Additional globals
@jinja_global("raise")
def jinja_raise(msg):
    raise TemplateError(msg)


# Additional tests
@jinja_test("match")
def test_match(txt, rgx, ignorecase=False, multiline=False):
    """Returns true if a sequence of chars matches a pattern."""
    flag = 0
    if ignorecase:
        flag |= re.I
    if multiline:
        flag |= re.M
    compiled_rgx = re.compile(rgx, flag)
    return True if compiled_rgx.match(txt) else False


@jinja_test("equalto")
def test_equalto(value, other):
    """Returns true if two values are equal."""
    return value == other


# Additional filters
@jinja_filter("skip")
def skip_filter(data):
    """
    Suppress data output

    .. code-block:: yaml

        {% my_string = "foo" %}

        {{ my_string|skip }}

    will be rendered as empty string,

    """
    return ""


@jinja_filter("sequence")
def ensure_sequence_filter(data):
    """
    Ensure sequenced data.

    **sequence**

        ensure that parsed data is a sequence

    .. code-block:: jinja

        {% set my_string = "foo" %}
        {% set my_list = ["bar", ] %}
        {% set my_dict = {"baz": "qux"} %}

        {{ my_string|sequence|first }}
        {{ my_list|sequence|first }}
        {{ my_dict|sequence|first }}


    will be rendered as:

    .. code-block:: yaml

        foo
        bar
        baz
    """
    if not isinstance(data, (list, tuple, set, dict)):
        return [data]
    return data


@jinja_filter("to_bool")
def to_bool(val):
    """
    Returns the logical value.

    .. code-block:: jinja

        {{ 'yes' | to_bool }}

    will be rendered as:

    .. code-block:: text

        True
    """
    if val is None:
        return False
    if isinstance(val, bool):
        return val
    if isinstance(val, (str, (str,))):
        return val.lower() in ("yes", "1", "true")
    if isinstance(val, int):
        return val > 0
    if not isinstance(val, Hashable):
        return len(val) > 0
    return False


@jinja_filter("indent")
def indent(s, width=4, first=False, blank=False, indentfirst=None):
    """
    A ported version of the "indent" filter containing a fix for indenting Markup
    objects. If the minion has Jinja version 2.11 or newer, the "indent" filter
    from upstream will be used, and this one will be ignored.
    """
    if indentfirst is not None:
        warnings.warn(
            "The 'indentfirst' argument is renamed to 'first' and will"
            " be removed in Jinja 3.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        first = indentfirst

    indention = " " * width
    newline = "\n"

    if isinstance(s, Markup):
        indention = Markup(indention)
        newline = Markup(newline)

    s += newline  # this quirk is necessary for splitlines method

    if blank:
        rv = (newline + indention).join(s.splitlines())
    else:
        lines = s.splitlines()
        rv = lines.pop(0)

        if lines:
            rv += newline + newline.join(
                indention + line if line else line for line in lines
            )

    if first:
        rv = indention + rv

    return rv


@jinja_filter("tojson")
def tojson(val, indent=None, **options):
    """
    Implementation of tojson filter (only present in Jinja 2.9 and later).
    Unlike the Jinja built-in filter, this allows arbitrary options to be
    passed in to the underlying JSON library.
    """
    options.setdefault("ensure_ascii", True)
    if indent is not None:
        options["indent"] = indent
    return (
        salt.utils.json.dumps(val, **options)
        .replace("<", "\\u003c")
        .replace(">", "\\u003e")
        .replace("&", "\\u0026")
        .replace("'", "\\u0027")
    )


@jinja_filter("quote")
def quote(txt):
    """
    Wraps a text around quotes.

    .. code-block:: jinja

        {% set my_text = 'my_text' %}
        {{ my_text | quote }}

    will be rendered as:

    .. code-block:: text

        'my_text'
    """
    return pipes.quote(txt)


@jinja_filter()
def regex_escape(value):
    return re.escape(value)


@jinja_filter("regex_search")
def regex_search(txt, rgx, ignorecase=False, multiline=False):
    """
    Searches for a pattern in the text.

    .. code-block:: jinja

        {% set my_text = 'abcd' %}
        {{ my_text | regex_search('^(.*)BC(.*)$', ignorecase=True) }}

    will be rendered as:

    .. code-block:: text

        ('a', 'd')
    """
    flag = 0
    if ignorecase:
        flag |= re.I
    if multiline:
        flag |= re.M
    obj = re.search(rgx, txt, flag)
    if not obj:
        return
    return obj.groups()


@jinja_filter("regex_match")
def regex_match(txt, rgx, ignorecase=False, multiline=False):
    """
    Searches for a pattern in the text.

    .. code-block:: jinja

        {% set my_text = 'abcd' %}
        {{ my_text | regex_match('^(.*)BC(.*)$', ignorecase=True) }}

    will be rendered as:

    .. code-block:: text

        ('a', 'd')
    """
    flag = 0
    if ignorecase:
        flag |= re.I
    if multiline:
        flag |= re.M
    obj = re.match(rgx, txt, flag)
    if not obj:
        return
    return obj.groups()


@jinja_filter("regex_replace")
def regex_replace(txt, rgx, val, ignorecase=False, multiline=False):
    r"""
    Searches for a pattern and replaces with a sequence of characters.

    .. code-block:: jinja

        {% set my_text = 'lets replace spaces' %}
        {{ my_text | regex_replace('\s+', '__') }}

    will be rendered as:

    .. code-block:: text

        lets__replace__spaces
    """
    flag = 0
    if ignorecase:
        flag |= re.I
    if multiline:
        flag |= re.M
    compiled_rgx = re.compile(rgx, flag)
    return compiled_rgx.sub(val, txt)


@jinja_filter("uuid")
def uuid_(val):
    """
    Returns a UUID corresponding to the value passed as argument.

    .. code-block:: jinja

        {{ 'example' | uuid }}

    will be rendered as:

    .. code-block:: text

        f4efeff8-c219-578a-bad7-3dc280612ec8
    """
    return str(uuid.uuid5(GLOBAL_UUID, salt.utils.stringutils.to_str(val)))


### List-related filters


@jinja_filter()
def unique(values):
    """
    Removes duplicates from a list.

    .. code-block:: jinja

        {% set my_list = ['a', 'b', 'c', 'a', 'b'] -%}
        {{ my_list | unique }}

    will be rendered as:

    .. code-block:: text

        ['a', 'b', 'c']
    """
    ret = None
    if isinstance(values, Hashable):
        ret = set(values)
    else:
        ret = []
        for value in values:
            if value not in ret:
                ret.append(value)
    return ret


@jinja_filter("min")
def lst_min(obj):
    """
    Returns the min value.

    .. code-block:: jinja

        {% set my_list = [1,2,3,4] -%}
        {{ my_list | min }}

    will be rendered as:

    .. code-block:: text

        1
    """
    return min(obj)


@jinja_filter("max")
def lst_max(obj):
    """
    Returns the max value.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | max }}

    will be rendered as:

    .. code-block:: text

        4
    """
    return max(obj)


@jinja_filter("avg")
def lst_avg(lst):
    """
    Returns the average value of a list.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | avg }}

    will be rendered as:

    .. code-block:: yaml

        2.5
    """
    if not isinstance(lst, Hashable):
        return float(sum(lst) / len(lst))
    return float(lst)


@jinja_filter("union")
def union(lst1, lst2):
    """
    Returns the union of two lists.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | union([2, 4, 6]) }}

    will be rendered as:

    .. code-block:: text

        [1, 2, 3, 4, 6]
    """
    if isinstance(lst1, Hashable) and isinstance(lst2, Hashable):
        return set(lst1) | set(lst2)
    return unique(lst1 + lst2)


@jinja_filter("intersect")
def intersect(lst1, lst2):
    """
    Returns the intersection of two lists.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | intersect([2, 4, 6]) }}

    will be rendered as:

    .. code-block:: text

        [2, 4]
    """
    if isinstance(lst1, Hashable) and isinstance(lst2, Hashable):
        return set(lst1) & set(lst2)
    return unique([ele for ele in lst1 if ele in lst2])


@jinja_filter("difference")
def difference(lst1, lst2):
    """
    Returns the difference of two lists.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | difference([2, 4, 6]) }}

    will be rendered as:

    .. code-block:: text

        [1, 3, 6]
    """
    if isinstance(lst1, Hashable) and isinstance(lst2, Hashable):
        return set(lst1) - set(lst2)
    return unique([ele for ele in lst1 if ele not in lst2])


@jinja_filter("symmetric_difference")
def symmetric_difference(lst1, lst2):
    """
    Returns the symmetric difference of two lists.

    .. code-block:: jinja

        {% my_list = [1,2,3,4] -%}
        {{ set my_list | symmetric_difference([2, 4, 6]) }}

    will be rendered as:

    .. code-block:: text

        [1, 3]
    """
    if isinstance(lst1, Hashable) and isinstance(lst2, Hashable):
        return set(lst1) ^ set(lst2)
    return unique(
        [ele for ele in union(lst1, lst2) if ele not in intersect(lst1, lst2)]
    )


@jinja_filter("method_call")
def method_call(obj, f_name, *f_args, **f_kwargs):
    return getattr(obj, f_name, lambda *args, **kwargs: None)(*f_args, **f_kwargs)


@jinja2.contextfunction
def show_full_context(ctx):
    return salt.utils.data.simple_types_filter(
        {key: value for key, value in ctx.items()}
    )


class SerializerExtension(Extension):
    '''
    Yaml and Json manipulation.

    **Format filters**

    Allows jsonifying or yamlifying any data structure. For example, this dataset:

    .. code-block:: python

        data = {
            'foo': True,
            'bar': 42,
            'baz': [1, 2, 3],
            'qux': 2.0
        }

    .. code-block:: jinja

        yaml = {{ data|yaml }}
        json = {{ data|json }}
        python = {{ data|python }}
        xml  = {{ {'root_node': data}|xml }}

    will be rendered as::

        yaml = {bar: 42, baz: [1, 2, 3], foo: true, qux: 2.0}
        json = {"baz": [1, 2, 3], "foo": true, "bar": 42, "qux": 2.0}
        python = {'bar': 42, 'baz': [1, 2, 3], 'foo': True, 'qux': 2.0}
        xml = """<<?xml version="1.0" ?>
                 <root_node bar="42" foo="True" qux="2.0">
                  <baz>1</baz>
                  <baz>2</baz>
                  <baz>3</baz>
                 </root_node>"""

    The yaml filter takes an optional flow_style parameter to control the
    default-flow-style parameter of the YAML dumper.

    .. code-block:: jinja

        {{ data|yaml(False) }}

    will be rendered as:

    .. code-block:: yaml

        bar: 42
        baz:
          - 1
          - 2
          - 3
        foo: true
        qux: 2.0

    **Load filters**

    Strings and variables can be deserialized with **load_yaml** and
    **load_json** tags and filters. It allows one to manipulate data directly
    in templates, easily:

    .. code-block:: jinja

        {%- set yaml_src = "{foo: it works}"|load_yaml %}
        {%- set json_src = "{'bar': 'for real'}"|load_json %}
        Dude, {{ yaml_src.foo }} {{ json_src.bar }}!

    will be rendered as::

        Dude, it works for real!

    **Load tags**

    Salt implements ``load_yaml`` and ``load_json`` tags. They work like
    the `import tag`_, except that the document is also deserialized.

    Syntaxes are ``{% load_yaml as [VARIABLE] %}[YOUR DATA]{% endload %}``
    and ``{% load_json as [VARIABLE] %}[YOUR DATA]{% endload %}``

    For example:

    .. code-block:: jinja

        {% load_yaml as yaml_src %}
            foo: it works
        {% endload %}
        {% load_json as json_src %}
            {
                "bar": "for real"
            }
        {% endload %}
        Dude, {{ yaml_src.foo }} {{ json_src.bar }}!

    will be rendered as::

        Dude, it works for real!

    **Import tags**

    External files can be imported and made available as a Jinja variable.

    .. code-block:: jinja

        {% import_yaml "myfile.yml" as myfile %}
        {% import_json "defaults.json" as defaults %}
        {% import_text "completeworksofshakespeare.txt" as poems %}

    **Catalog**

    ``import_*`` and ``load_*`` tags will automatically expose their
    target variable to import. This feature makes catalog of data to
    handle.

    for example:

    .. code-block:: jinja

        # doc1.sls
        {% load_yaml as var1 %}
            foo: it works
        {% endload %}
        {% load_yaml as var2 %}
            bar: for real
        {% endload %}

    .. code-block:: jinja

        # doc2.sls
        {% from "doc1.sls" import var1, var2 as local2 %}
        {{ var1.foo }} {{ local2.bar }}

    ** Escape Filters **

    .. versionadded:: 2017.7.0

    Allows escaping of strings so they can be interpreted literally by another
    function.

    For example:

    .. code-block:: jinja

        regex_escape = {{ 'https://example.com?foo=bar%20baz' | regex_escape }}

    will be rendered as::

        regex_escape = https\\:\\/\\/example\\.com\\?foo\\=bar\\%20baz

    ** Set Theory Filters **

    .. versionadded:: 2017.7.0

    Performs set math using Jinja filters.

    For example:

    .. code-block:: jinja

        unique = {{ ['foo', 'foo', 'bar'] | unique }}

    will be rendered as::

        unique = ['foo', 'bar']

    .. _`import tag`: https://jinja.palletsprojects.com/en/2.11.x/templates/#import
    '''

    tags = {
        "load_yaml",
        "load_json",
        "import_yaml",
        "import_json",
        "load_text",
        "import_text",
        "profile",
    }

    def __init__(self, environment):
        super().__init__(environment)
        self.environment.filters.update(
            {
                "yaml": self.format_yaml,
                "json": self.format_json,
                "xml": self.format_xml,
                "python": self.format_python,
                "load_yaml": self.load_yaml,
                "load_json": self.load_json,
                "load_text": self.load_text,
            }
        )

        if self.environment.finalize is None:
            self.environment.finalize = self.finalizer
        else:
            finalizer = self.environment.finalize

            @wraps(finalizer)
            def wrapper(self, data):
                return finalizer(self.finalizer(data))

            self.environment.finalize = wrapper

    def finalizer(self, data):
        """
        Ensure that printed mappings are YAML friendly.
        """

        def explore(data):
            if isinstance(data, (dict, OrderedDict)):
                return PrintableDict(
                    [(key, explore(value)) for key, value in data.items()]
                )
            elif isinstance(data, (list, tuple, set)):
                return data.__class__([explore(value) for value in data])
            return data

        return explore(data)

    def format_json(self, value, sort_keys=True, indent=None):
        json_txt = salt.utils.json.dumps(
            value, sort_keys=sort_keys, indent=indent
        ).strip()
        try:
            return Markup(json_txt)
        except UnicodeDecodeError:
            return Markup(salt.utils.stringutils.to_unicode(json_txt))

    def format_yaml(self, value, flow_style=True):
        yaml_txt = salt.utils.yaml.safe_dump(
            value, default_flow_style=flow_style
        ).strip()
        if yaml_txt.endswith("\n..."):
            yaml_txt = yaml_txt[: len(yaml_txt) - 4]
        try:
            return Markup(yaml_txt)
        except UnicodeDecodeError:
            return Markup(salt.utils.stringutils.to_unicode(yaml_txt))

    def format_xml(self, value):
        """Render a formatted multi-line XML string from a complex Python
        data structure. Supports tag attributes and nested dicts/lists.

        :param value: Complex data structure representing XML contents
        :returns: Formatted XML string rendered with newlines and indentation
        :rtype: str
        """

        def normalize_iter(value):
            if isinstance(value, (list, tuple)):
                if isinstance(value[0], str):
                    xmlval = value
                else:
                    xmlval = []
            elif isinstance(value, dict):
                xmlval = list(value.items())
            else:
                raise TemplateRuntimeError(
                    "Value is not a dict or list. Cannot render as XML"
                )
            return xmlval

        def recurse_tree(xmliter, element=None):
            sub = None
            for tag, attrs in xmliter:
                if isinstance(attrs, list):
                    for attr in attrs:
                        recurse_tree(((tag, attr),), element)
                elif element is not None:
                    sub = SubElement(element, tag)
                else:
                    sub = Element(tag)
                if isinstance(attrs, (str, int, bool, float)):
                    sub.text = str(attrs)
                    continue
                if isinstance(attrs, dict):
                    sub.attrib = {
                        attr: str(val)
                        for attr, val in attrs.items()
                        if not isinstance(val, (dict, list))
                    }
                for tag, val in [
                    item
                    for item in normalize_iter(attrs)
                    if isinstance(item[1], (dict, list))
                ]:
                    recurse_tree(((tag, val),), sub)
            return sub

        return Markup(
            minidom.parseString(
                tostring(recurse_tree(normalize_iter(value)))
            ).toprettyxml(indent=" ")
        )

    def format_python(self, value):
        return Markup(pprint.pformat(value).strip())

    def load_yaml(self, value):
        if isinstance(value, TemplateModule):
            value = str(value)
        try:
            return salt.utils.data.decode(salt.utils.yaml.safe_load(value))
        except salt.utils.yaml.YAMLError as exc:
            msg = "Encountered error loading yaml: "
            try:
                # Reported line is off by one, add 1 to correct it
                line = exc.problem_mark.line + 1
                buf = exc.problem_mark.buffer
                problem = exc.problem
            except AttributeError:
                # No context information available in the exception, fall back
                # to the stringified version of the exception.
                msg += str(exc)
            else:
                msg += "{}\n".format(problem)
                msg += salt.utils.stringutils.get_context(
                    buf, line, marker="    <======================"
                )
            raise TemplateRuntimeError(msg)
        except AttributeError:
            raise TemplateRuntimeError("Unable to load yaml from {}".format(value))

    def load_json(self, value):
        if isinstance(value, TemplateModule):
            value = str(value)
        try:
            return salt.utils.json.loads(value)
        except (ValueError, TypeError, AttributeError):
            raise TemplateRuntimeError("Unable to load json from {}".format(value))

    def load_text(self, value):
        if isinstance(value, TemplateModule):
            value = str(value)

        return value

    _load_parsers = {"load_yaml", "load_json", "load_text"}
    _import_parsers = {"import_yaml", "import_json", "import_text"}

    def parse(self, parser):
        if parser.stream.current.value in self._load_parsers:
            return self.parse_load(parser)
        elif parser.stream.current.value in self._import_parsers:
            return self.parse_import(
                parser, parser.stream.current.value.split("_", 1)[1]
            )
        elif parser.stream.current.value == "profile":
            return self.parse_profile(parser)

        parser.fail(
            "Unknown format " + parser.stream.current.value,
            parser.stream.current.lineno,
        )

    # pylint: disable=E1120,E1121
    def parse_profile(self, parser):
        lineno = next(parser.stream).lineno
        parser.stream.expect("name:as")
        label = parser.parse_expression()
        body = parser.parse_statements(["name:endprofile"], drop_needle=True)
        return self._parse_profile_block(parser, label, "profile block", body, lineno)

    def _create_profile_id(self, parser):
        return "_salt_profile_{}".format(parser.free_identifier().name)

    def _profile_start(self, label, source):
        return (label, source, time.time())

    def _profile_end(self, label, source, previous_time):
        log.profile(
            "Time (in seconds) to render %s '%s': %s",
            source,
            label,
            time.time() - previous_time,
        )

    def _parse_profile_block(self, parser, label, source, body, lineno):
        profile_id = self._create_profile_id(parser)
        ret = (
            [
                nodes.Assign(
                    nodes.Name(profile_id, "store").set_lineno(lineno),
                    self.call_method(
                        "_profile_start",
                        dyn_args=nodes.List([label, nodes.Const(source)]).set_lineno(
                            lineno
                        ),
                    ).set_lineno(lineno),
                ).set_lineno(lineno),
            ]
            + body
            + [
                nodes.ExprStmt(
                    self.call_method(
                        "_profile_end", dyn_args=nodes.Name(profile_id, "load")
                    ),
                ).set_lineno(lineno),
            ]
        )
        return ret

    def parse_load(self, parser):
        filter_name = parser.stream.current.value
        lineno = next(parser.stream).lineno
        if filter_name not in self.environment.filters:
            parser.fail("Unable to parse {}".format(filter_name), lineno)

        parser.stream.expect("name:as")
        target = parser.parse_assign_target()
        macro_name = "_" + parser.free_identifier().name
        macro_body = parser.parse_statements(("name:endload",), drop_needle=True)

        return [
            nodes.Macro(macro_name, [], [], macro_body).set_lineno(lineno),
            nodes.Assign(
                target,
                nodes.Filter(
                    nodes.Call(
                        nodes.Name(macro_name, "load").set_lineno(lineno),
                        [],
                        [],
                        None,
                        None,
                    ).set_lineno(lineno),
                    filter_name,
                    [],
                    [],
                    None,
                    None,
                ).set_lineno(lineno),
            ).set_lineno(lineno),
        ]

    def parse_import(self, parser, converter):
        import_node = parser.parse_import()
        target = import_node.target
        lineno = import_node.lineno

        body = [
            import_node,
            nodes.Assign(
                nodes.Name(target, "store").set_lineno(lineno),
                nodes.Filter(
                    nodes.Name(target, "load").set_lineno(lineno),
                    "load_{}".format(converter),
                    [],
                    [],
                    None,
                    None,
                ).set_lineno(lineno),
            ).set_lineno(lineno),
        ]
        return self._parse_profile_block(
            parser, import_node.template, "import_{}".format(converter), body, lineno
        )

    # pylint: enable=E1120,E1121



r"""
Renderer that will decrypt GPG ciphers

Any key in the SLS file can be a GPG cipher, and this renderer will decrypt it
before passing it off to Salt. This allows you to safely store secrets in
source control, in such a way that only your Salt master can decrypt them and
distribute them only to the minions that need them.

The typical use-case would be to use ciphers in your pillar data, and keep a
secret key on your master. You can put the public key in source control so that
developers can add new secrets quickly and easily.

This renderer requires the gpg_ binary. No python libraries are required as of
the 2015.8.0 release.

.. _gpg-homedir:

GPG Homedir
-----------

When running gpg commands, it is important to run commands as the user that owns
the keys directory. If salt-master runs as user salt, then ``su salt`` before
running any gpg commands.

To avoid compatibility and upgrade problems and to provide a standardized location
for keys, salt uses ``/etc/salt/gpgkeys``. In order to make the gpg command use
this directory, use ``gpg --homedir /etc/salt/gpgkeys`` with gpg commands or set
the homedir for that user using ``echo 'homedir /etc/salt/gpgkeys' >> ~/.gnupg``.

.. _gpg: https://gnupg.org

Setup
-----

To set things up, first generate a keypair. On the master, run the following:

.. code-block:: bash

    # mkdir -p /etc/salt/gpgkeys
    # chmod 0700 /etc/salt/gpgkeys
    # gpg --gen-key --homedir /etc/salt/gpgkeys

Do not supply a password for the keypair, and use a name that makes sense for
your application. Be sure to back up the ``gpgkeys`` directory someplace safe!

.. note::
    Unfortunately, there are some scenarios - for example, on virtual machines
    which don’t have real hardware - where insufficient entropy causes key
    generation to be extremely slow. In these cases, there are usually means of
    increasing the system entropy. On virtualised Linux systems, this can often
    be achieved by installing the ``rng-tools`` package.

Import keys to a master
***********************

If the keys already exist and need to be imported to the salt master, run the
following to import them.

.. code-block:: bash

    gpg --homedir /etc/salt/gpgkeys --import /path/to/private.key
    gpg --homedir /etc/salt/gpgkeys --import /path/to/pubkey.gpg

Note: The default `GPG Homedir <gpg-homedir>` is ``~/.gnupg`` and needs to be
set using ``--homedir``.

Adjust trust level of imported keys
***********************************

In some cases, importing existing keys may not be enough and the trust level of
the key needs to be adjusted. This can be done by editing the key. The ``key_id``
and the actual trust level of the key can be seen by listing the already imported
keys.

.. code-block:: bash

    gpg --homedir /etc/salt/gpgkeys --list-keys
    gpg --homedir /etc/salt/gpgkeys --list-secret-keys

If the trust-level is not ``ultimate`` it needs to be changed by running

.. code-block:: bash

    gpg --homedir /etc/salt/gpgkeys --edit-key <key_id>

This will open an interactive shell for the management of the GPG encryption key.
Type ``trust`` to be able to set the trust level for the key and then select ``5
(I trust ultimately)``. Then quit the shell by typing ``save``.

Different GPG Location
**********************

In some cases, it's preferable to have gpg keys stored on removable media or
other non-standard locations. This can be done using the ``gpg_keydir`` option
on the salt master. This will also require using a different path to ``--homedir``,
as mentioned in the `GPG Homedir <gpg-homedir>` section.

.. code-block:: bash

    gpg_keydir: <path/to/homedir>

Export the Public Key
---------------------

.. code-block:: bash

    # gpg --homedir /etc/salt/gpgkeys --armor --export <KEY-NAME> > exported_pubkey.gpg


Import the Public Key
---------------------

To encrypt secrets, copy the public key to your local machine and run:

.. code-block:: bash

    $ gpg --import exported_pubkey.gpg

To generate a cipher from a secret:

.. code-block:: bash

   $ echo -n "supersecret" | gpg --armor --batch --trust-model always --encrypt -r <KEY-name>

To apply the renderer on a file-by-file basis add the following line to the
top of any pillar with gpg data in it:

.. code-block:: yaml

    #!yaml|gpg

Now with your renderer configured, you can include your ciphers in your pillar
data like so:

.. code-block:: yaml

    #!yaml|gpg

    a-secret: |
      -----BEGIN PGP MESSAGE-----
      Version: GnuPG v1

      hQEMAweRHKaPCfNeAQf9GLTN16hCfXAbPwU6BbBK0unOc7i9/etGuVc5CyU9Q6um
      QuetdvQVLFO/HkrC4lgeNQdM6D9E8PKonMlgJPyUvC8ggxhj0/IPFEKmrsnv2k6+
      cnEfmVexS7o/U1VOVjoyUeliMCJlAz/30RXaME49Cpi6No2+vKD8a4q4nZN1UZcG
      RhkhC0S22zNxOXQ38TBkmtJcqxnqT6YWKTUsjVubW3bVC+u2HGqJHu79wmwuN8tz
      m4wBkfCAd8Eyo2jEnWQcM4TcXiF01XPL4z4g1/9AAxh+Q4d8RIRP4fbw7ct4nCJv
      Gr9v2DTF7HNigIMl4ivMIn9fp+EZurJNiQskLgNbktJGAeEKYkqX5iCuB1b693hJ
      FKlwHiJt5yA8X2dDtfk8/Ph1Jx2TwGS+lGjlZaNqp3R1xuAZzXzZMLyZDe5+i3RJ
      skqmFTbOiA===Eqsm
      -----END PGP MESSAGE-----


.. _encrypted-cli-pillar-data:

Encrypted CLI Pillar Data
-------------------------

.. versionadded:: 2016.3.0

Functions like :py:func:`state.highstate <salt.modules.state.highstate>` and
:py:func:`state.sls <salt.modules.state.sls>` allow for pillar data to be
passed on the CLI.

.. code-block:: bash

    salt myminion state.highstate pillar="{'mypillar': 'foo'}"

Starting with the 2016.3.0 release of Salt, it is now possible for this pillar
data to be GPG-encrypted, and to use the GPG renderer to decrypt it.


Replacing Newlines
******************

To pass encrypted pillar data on the CLI, the ciphertext must have its newlines
replaced with a literal backslash-n (``\n``), as newlines are not supported
within Salt CLI arguments. There are a number of ways to do this:

With awk or Perl:

.. code-block:: bash

    # awk
    ciphertext=`echo -n "supersecret" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | awk '{printf "%s\\n",$0} END {print ""}'`
    # Perl
    ciphertext=`echo -n "supersecret" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | perl -pe 's/\n/\\n/g'`

With Python:

.. code-block:: python

    import subprocess

    secret, stderr = subprocess.Popen(
        ['gpg', '--armor', '--batch', '--trust-model', 'always', '--encrypt',
         '-r', 'user@domain.com'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE).communicate(input='supersecret')

    if secret:
        print(secret.replace('\n', r'\n'))
    else:
        raise ValueError('No ciphertext found: {0}'.format(stderr))

.. code-block:: bash

    ciphertext=`python /path/to/script.py`


The ciphertext can be included in the CLI pillar data like so:

.. code-block:: bash

    salt myminion state.sls secretstuff pillar_enc=gpg pillar="{secret_pillar: '$ciphertext'}"

The ``pillar_enc=gpg`` argument tells Salt that there is GPG-encrypted pillar
data, so that the CLI pillar data is passed through the GPG renderer, which
will iterate recursively though the CLI pillar dictionary to decrypt any
encrypted values.


Encrypting the Entire CLI Pillar Dictionary
*******************************************

If several values need to be encrypted, it may be more convenient to encrypt
the entire CLI pillar dictionary. Again, this can be done in several ways:

With awk or Perl:

.. code-block:: bash

    # awk
    ciphertext=`echo -n "{'secret_a': 'CorrectHorseBatteryStaple', 'secret_b': 'GPG is fun!'}" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | awk '{printf "%s\\n",$0} END {print ""}'`
    # Perl
    ciphertext=`echo -n "{'secret_a': 'CorrectHorseBatteryStaple', 'secret_b': 'GPG is fun!'}" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | perl -pe 's/\n/\\n/g'`

With Python:

.. code-block:: python

    import subprocess

    pillar_data = {'secret_a': 'CorrectHorseBatteryStaple',
                   'secret_b': 'GPG is fun!'}

    secret, stderr = subprocess.Popen(
        ['gpg', '--armor', '--batch', '--trust-model', 'always', '--encrypt',
         '-r', 'user@domain.com'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE).communicate(input=repr(pillar_data))

    if secret:
        print(secret.replace('\n', r'\n'))
    else:
        raise ValueError('No ciphertext found: {0}'.format(stderr))

.. code-block:: bash

    ciphertext=`python /path/to/script.py`

With the entire pillar dictionary now encrypted, it can be included in the CLI
pillar data like so:

.. code-block:: bash

    salt myminion state.sls secretstuff pillar_enc=gpg pillar="$ciphertext"
"""


import logging
import os
import re
from subprocess import PIPE, Popen

import salt.syspaths
import salt.utils.cache
import salt.utils.path
import salt.utils.stringio
import salt.utils.stringutils
from salt.exceptions import SaltRenderError

log = logging.getLogger(__name__)

GPG_CIPHERTEXT = re.compile(
    salt.utils.stringutils.to_bytes(
        r"-----BEGIN PGP MESSAGE-----.*?-----END PGP MESSAGE-----"
    ),
    re.DOTALL,
)
GPG_CACHE = None


def _get_gpg_exec():
    """
    return the GPG executable or raise an error
    """
    gpg_exec = salt.utils.path.which("gpg")
    if gpg_exec:
        return gpg_exec
    else:
        raise SaltRenderError("GPG unavailable")


def _get_key_dir():
    """
    return the location of the GPG key directory
    """
    gpg_keydir = None
    if "config.get" in __salt__:
        gpg_keydir = __salt__["config.get"]("gpg_keydir")

    if not gpg_keydir:
        gpg_keydir = __opts__.get(
            "gpg_keydir",
            os.path.join(
                __opts__.get("config_dir", os.path.dirname(__opts__["conf_file"])),
                "gpgkeys",
            ),
        )

    return gpg_keydir


def _get_cache():
    global GPG_CACHE
    if not GPG_CACHE:
        cachedir = __opts__.get("cachedir")
        GPG_CACHE = salt.utils.cache.CacheFactory.factory(
            __opts__.get("gpg_cache_backend"),
            __opts__.get("gpg_cache_ttl"),
            minion_cache_path=os.path.join(cachedir, "gpg_cache"),
        )
    return GPG_CACHE


def _decrypt_ciphertext(cipher):
    """
    Given a block of ciphertext as a string, and a gpg object, try to decrypt
    the cipher and return the decrypted string. If the cipher cannot be
    decrypted, log the error, and return the ciphertext back out.
    """
    try:
        cipher = salt.utils.stringutils.to_unicode(cipher).replace(r"\n", "\n")
    except UnicodeDecodeError:
        # ciphertext is binary
        pass
    cipher = salt.utils.stringutils.to_bytes(cipher)
    if __opts__.get("gpg_cache"):
        cache = _get_cache()
        if cipher in cache:
            return cache[cipher]
    cmd = [
        _get_gpg_exec(),
        "--homedir",
        _get_key_dir(),
        "--status-fd",
        "2",
        "--no-tty",
        "-d",
    ]
    proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=False)
    decrypted_data, decrypt_error = proc.communicate(input=cipher)
    if not decrypted_data:
        log.warning("Could not decrypt cipher %r, received: %r", cipher, decrypt_error)
        return cipher
    else:
        if __opts__.get("gpg_cache"):
            cache[cipher] = decrypted_data
        return decrypted_data


def _decrypt_ciphertexts(cipher, translate_newlines=False, encoding=None):
    to_bytes = salt.utils.stringutils.to_bytes
    cipher = to_bytes(cipher)
    if translate_newlines:
        cipher = cipher.replace(to_bytes(r"\n"), to_bytes("\n"))

    def replace(match):
        result = to_bytes(_decrypt_ciphertext(match.group()))
        return result

    ret, num = GPG_CIPHERTEXT.subn(replace, to_bytes(cipher))
    if num > 0:
        # Remove trailing newlines. Without if crypted value initially specified as a YAML multiline
        # it will conain unexpected trailing newline.
        ret = ret.rstrip(b"\n")
    else:
        ret = cipher

    try:
        ret = salt.utils.stringutils.to_unicode(ret, encoding=encoding)
    except UnicodeDecodeError:
        # decrypted data contains some sort of binary data - not our problem
        pass
    return ret


def _decrypt_object(obj, translate_newlines=False, encoding=None):
    """
    Recursively try to decrypt any object. If the object is a string
    or bytes and it contains a valid GPG header, decrypt it,
    otherwise keep going until a string is found.
    """
    if salt.utils.stringio.is_readable(obj):
        return _decrypt_object(obj.getvalue(), translate_newlines)
    if isinstance(obj, (str, bytes)):
        return _decrypt_ciphertexts(
            obj, translate_newlines=translate_newlines, encoding=encoding
        )
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = _decrypt_object(value, translate_newlines=translate_newlines)
        return obj
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = _decrypt_object(value, translate_newlines=translate_newlines)
        return obj
    else:
        return obj


def render(gpg_data, saltenv="base", sls="", argline="", **kwargs):
    """
    Create a gpg object given a gpg_keydir, and then use it to try to decrypt
    the data to be rendered.
    """
    if not _get_gpg_exec():
        raise SaltRenderError("GPG unavailable")
    log.debug("Reading GPG keys from: %s", _get_key_dir())

    translate_newlines = kwargs.get("translate_newlines", False)
    return _decrypt_object(
        gpg_data,
        translate_newlines=translate_newlines,
        encoding=kwargs.get("encoding", None),
    )



"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import salt.proxy.napalm as napalm_proxy
import tests.support.napalm as napalm_test_support
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NapalmProxyTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        def mock_get_device(opts, *args, **kwargs):
            assert opts == self.test_opts
            return {"DRIVER": napalm_test_support.MockNapalmDevice(), "UP": True}

        patcher = patch("salt.utils.napalm.get_device", mock_get_device)
        patcher.start()
        self.addCleanup(patcher.stop)
        self.test_opts = {
            "proxytype": "napalm",
            "driver": "junos",
            "host": "core05.nrt02",
            "id": "core05.nrt02",
        }
        module_globals = {
            "__salt__": {
                "config.option": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                )
            }
        }
        module_globals["napalm_base"] = MagicMock()
        return {napalm_proxy: module_globals}

    def test_init(self):
        ret = napalm_proxy.init(self.test_opts)
        assert ret is True

    def test_alive(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.alive(self.test_opts)
        assert ret is True

    def test_ping(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.ping()
        assert ret is True

    def test_initialized(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.initialized()
        assert ret is True

    def test_get_device(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.get_device()
        assert ret["UP"] is True

    def test_get_grains(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.get_grains()
        assert ret["out"] == napalm_test_support.TEST_FACTS.copy()

    def test_grains_refresh(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.grains_refresh()
        assert ret["out"] == napalm_test_support.TEST_FACTS.copy()

    def test_fns(self):
        ret = napalm_proxy.fns()
        assert "details" in ret.keys()

    def test_shutdown(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.shutdown(self.test_opts)
        assert ret is True

    def test_call(self):
        napalm_proxy.init(self.test_opts)
        ret = napalm_proxy.call("get_arp_table")
        assert ret["out"] == napalm_test_support.TEST_ARP_TABLE.copy()



"""
    :codeauthor: Jorge Schrauwen <info@blackdot.be>
"""


import salt.modules.smartos_imgadm as imgadm
from salt.modules.smartos_imgadm import _parse_image_meta
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase

image_orphan = {
    "manifest": {"uuid": "07f360fd-12d5-e624-a279-eb8a15b630f6"},
    "zpool": "zones",
    "cloneNames": [],
    "clones": 0,
}
image_native = {
    "manifest": {
        "v": 2,
        "uuid": "9d91e334-3bcf-11e8-bb0b-e7b49eb01e38",
        "owner": "00000000-0000-0000-0000-000000000000",
        "name": "pkgbuild",
        "version": "18.1.0",
        "state": "active",
        "disabled": False,
        "public": True,
        "published_at": "2018-04-09T08:25:52Z",
        "type": "zone-dataset",
        "os": "smartos",
        "files": [
            {
                "sha1": "5efaf95b7f226eb09c7d5e6c3734f8aa654b811d",
                "size": 465411979,
                "compression": "gzip",
            }
        ],
        "description": "A SmartOS image pre-configured for building pkgsrc packages.",
        "homepage": "https://docs.joyent.com/images/smartos/pkgbuild",
        "urn": "sdc:sdc:pkgbuild:18.1.0",
        "requirements": {
            "min_platform": {"7.0": "20141030T081701Z"},
            "networks": [{"name": "net0", "description": "public"}],
        },
        "tags": {"role": "os", "group": "pkgbuild"},
    },
    "zpool": "zones",
    "source": "https://images.joyent.com",
    "cloneNames": ["zones/dda70f61-70fe-65e7-cf70-d878d69442d4"],
    "clones": 1,
}
image_lx = {
    "manifest": {
        "v": 2,
        "uuid": "05140a7e-279f-11e6-aedf-47d4f69d2887",
        "owner": "00000000-0000-0000-0000-000000000000",
        "name": "ubuntu-16.04",
        "version": "20160601",
        "state": "active",
        "disabled": False,
        "public": True,
        "published_at": "2016-06-01T02:17:41Z",
        "type": "lx-dataset",
        "os": "linux",
        "files": [
            {
                "sha1": "d342f137c5ccef0702ec479acb63c196cf81b38a",
                "size": 134969110,
                "compression": "gzip",
            }
        ],
        "description": (
            "Container-native Ubuntu 16.04 64-bit image. Built to run on containers"
            " with bare metal speed, while offering all the services of a typical unix"
            " host."
        ),
        "homepage": "https://docs.joyent.com/images/container-native-linux",
        "requirements": {
            "networks": [{"name": "net0", "description": "public"}],
            "min_platform": {"7.0": "20160225T122859Z"},
            "brand": "lx",
        },
        "tags": {"role": "os", "kernel_version": "4.3.0"},
    },
    "zpool": "zones",
    "source": "https://images.joyent.com",
    "cloneNames": ["zones/e4c1f6b5-4429-e6c2-ae2a-d6aa58bdeebb"],
    "clones": 1,
}
image_zvol = {
    "manifest": {
        "v": 2,
        "uuid": "ac99517a-72ac-44c0-90e6-c7ce3d944a0a",
        "owner": "00000000-0000-0000-0000-000000000000",
        "name": "ubuntu-certified-18.04",
        "version": "20180808",
        "state": "active",
        "disabled": False,
        "public": True,
        "published_at": "2018-10-11T12:45:24.804Z",
        "type": "zvol",
        "os": "linux",
        "files": [
            {
                "sha1": "9f7704969507bd97e160a8f42a3631487644e457",
                "size": 372276887,
                "compression": "gzip",
            }
        ],
        "description": (
            "Ubuntu 18.04 LTS (20180808 64-bit). Certified Ubuntu Server Cloud Image"
            " from Canonical. For kvm and bhyve."
        ),
        "homepage": "https://docs.joyent.com/images/linux/ubuntu-certified",
        "requirements": {
            "min_platform": {"7.0": "20150929T232348Z"},
            "networks": [{"name": "net0", "description": "public"}],
            "ssh_key": True,
        },
        "nic_driver": "virtio",
        "disk_driver": "virtio",
        "cpu_type": "host",
        "image_size": 10240,
        "tags": {"default_user": "ubuntu", "role": "os"},
    },
    "zpool": "zones",
    "source": "https://images.joyent.com",
    "cloneNames": [],
    "clones": 0,
}
image_docker = {
    "manifest": {
        "v": 2,
        "uuid": "4a3db8cb-0e94-ae23-588c-ee7934088927",
        "owner": "00000000-0000-0000-0000-000000000000",
        "name": "docker-layer",
        "version": "62487cf6a7f6",
        "disabled": False,
        "public": True,
        "published_at": "2019-03-23T01:32:25.320Z",
        "type": "docker",
        "os": "linux",
        "description": '/bin/sh -c #(nop)  CMD ["/bin/bash" "/opt/start.sh" "-bash"]',
        "tags": {
            "docker:repo": "busybox42/zimbra-docker-centos",
            "docker:id": "sha256:62487cf6a7f698af4edc20707e14b1b3bba13b98bea3375f05af04859a30b222",
            "docker:architecture": "amd64",
            "docker:tag:latest": True,
            "docker:config": {
                "Cmd": ["/bin/bash", "/opt/start.sh", "-bash"],
                "Entrypoint": None,
                "Env": [
                    "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
                ],
                "WorkingDir": "",
            },
        },
        "origin": "2f0c529b-7bab-28d1-ff34-bdc9281b7a4b",
    },
    "zpool": "zones",
    "source": "https://docker.io",
    "cloneNames": [],
    "clones": 0,
}


class ImgadmTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.smartos_imgadm module
    """

    def setup_loader_modules(self):
        return {imgadm: {}}

    def test_parse_image_meta_orphan(self):
        """
        Test the internal _parse_image_meta methode
        Feed it an 'orphan' image as we get it from from imgadm list -j
        """
        ret = {"Error": "This looks like an orphaned image, image payload was invalid."}
        self.assertEqual(_parse_image_meta(image_orphan, True), ret)

    def test_parse_image_meta_native(self):
        """
        Test the internal _parse_image_meta methode
        Feed it an 'native' image as we get it from from imgadm list -j
        """
        ret = {
            "description": (
                "A SmartOS image pre-configured for building pkgsrc packages."
            ),
            "name": "pkgbuild",
            "os": "smartos",
            "published": "2018-04-09T08:25:52Z",
            "source": "https://images.joyent.com",
            "version": "18.1.0",
        }
        self.assertEqual(_parse_image_meta(image_native, True), ret)

    def test_parse_image_meta_lx(self):
        """
        Test the internal _parse_image_meta methode
        Feed it an 'lx' image as we get it from from imgadm list -j
        """
        ret = {
            "description": (
                "Container-native Ubuntu 16.04 64-bit image. Built to run on "
                "containers with bare metal speed, while offering all the "
                "services of a typical unix host."
            ),
            "name": "ubuntu-16.04",
            "os": "linux",
            "published": "2016-06-01T02:17:41Z",
            "source": "https://images.joyent.com",
            "version": "20160601",
        }
        self.assertEqual(_parse_image_meta(image_lx, True), ret)

    def test_parse_image_meta_zvol(self):
        """
        Test the internal _parse_image_meta methode
        Feed it an 'zvol' image as we get it from from imgadm list -j
        """
        ret = {
            "description": (
                "Ubuntu 18.04 LTS (20180808 64-bit). Certified Ubuntu Server "
                "Cloud Image from Canonical. For kvm and bhyve."
            ),
            "name": "ubuntu-certified-18.04",
            "os": "linux",
            "published": "2018-10-11T12:45:24.804Z",
            "source": "https://images.joyent.com",
            "version": "20180808",
        }
        self.assertEqual(_parse_image_meta(image_zvol, True), ret)

    def test_parse_image_meta_docker(self):
        """
        Test the internal _parse_image_meta methode
        Feed it an 'docker' image as we get it from from imgadm list -j
        """
        ret = {
            "description": (
                "Docker image imported from "
                "busybox42/zimbra-docker-centos:latest on "
                "2019-03-23T01:32:25.320Z."
            ),
            "name": "busybox42/zimbra-docker-centos:latest",
            "os": "linux",
            "published": "2019-03-23T01:32:25.320Z",
            "source": "https://docker.io",
            "version": "62487cf6a7f6",
        }
        self.assertEqual(_parse_image_meta(image_docker, True), ret)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.varnish as varnish
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class VarnishTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.varnish
    """

    def setup_loader_modules(self):
        return {varnish: {}}

    def test_version(self):
        """
        Test to return server version from varnishd -V
        """
        with patch.dict(
            varnish.__salt__, {"cmd.run": MagicMock(return_value="(varnish-2.0)")}
        ):
            self.assertEqual(varnish.version(), "2.0")

    def test_ban(self):
        """
        Test to add ban to the varnish cache
        """
        with patch.object(varnish, "_run_varnishadm", return_value={"retcode": 0}):
            self.assertTrue(varnish.ban("ban_expression"))

    def test_ban_list(self):
        """
        Test to list varnish cache current bans
        """
        with patch.object(varnish, "_run_varnishadm", return_value={"retcode": True}):
            self.assertFalse(varnish.ban_list())

        with patch.object(
            varnish,
            "_run_varnishadm",
            return_value={"retcode": False, "stdout": "A\nB\nC"},
        ):
            self.assertEqual(varnish.ban_list(), ["B", "C"])

    def test_purge(self):
        """
        Test to purge the varnish cache
        """
        with patch.object(varnish, "ban", return_value=True):
            self.assertTrue(varnish.purge())

    def test_param_set(self):
        """
        Test to set a param in varnish cache
        """
        with patch.object(varnish, "_run_varnishadm", return_value={"retcode": 0}):
            self.assertTrue(varnish.param_set("param", "value"))

    def test_param_show(self):
        """
        Test to show params of varnish cache
        """
        with patch.object(
            varnish,
            "_run_varnishadm",
            return_value={"retcode": True, "stdout": "A\nB\nC"},
        ):
            self.assertFalse(varnish.param_show("param"))

        with patch.object(
            varnish,
            "_run_varnishadm",
            return_value={"retcode": False, "stdout": "A .1\nB .2\n"},
        ):
            self.assertEqual(varnish.param_show("param"), {"A": ".1"})
